#ABC = Abstract Base Class
from abc import ABC

class FacilityABC(ABC):
    def __init__(self, *, storage: object, plugin_manager: object, settings: object):
        self.storage = storage
        self.plugin_manager = plugin_manager
        self.settings = settings


import asyncio
import threading

class StageResponse:
    def __init__(self, stage, task, on_success=None, on_error=None):
        self._stage = stage
        self._stage._responses.add(self)
        self._loop = self._stage._loop
        self._on_success = on_success
        self._on_error = on_error
        if asyncio.iscoroutine(task):
            self._task = asyncio.run_coroutine_threadsafe(task, self._loop)
        elif asyncio.isfuture(task):
            self._task = task
        self._task.add_done_callback(self._on_task_done)
        self._result_ready = threading.Event()
        self._status = None
        self._result = None
        self._error = None
        self._final_result = None
    
    def _on_task_done(self, future):
        try:
            self._status = True
            self._result = future.result()
            if self._on_success:
                self._final_result = self._on_success(self._result)
            self._result_ready.set()
            self._stage._responses.discard(self)
        except Exception as e:
            self._status = False
            self._error = e
            if self._on_error:
                self._final_result = self._on_error(self._error)
                self._result_ready.set()
                self._stage._responses.discard(self)
            else:
                self._result_ready.set()
                self._stage._responses.discard(self)
                raise self._error
    
    def get(self):
        self._result_ready.wait()
        if self._status == True:
            return self._result
        elif self._on_error is None:
            raise self._error
    
    def get_final(self):
        self._result_ready.wait()
        if self._final_result:
            return self._final_result
        else:
            return self._result


from typing import Union
from Agently.Agent import Agent
from Agently.Workflow import Workflow
from mcp.server import FastMCP

class MCPServerHandler:
    def serve_agent(self, agent: Agent, *, name:str=None, desc:str=None):
        if name is None or name == "":
            name = "Agently Agent"
        if desc is None or desc == "":
            desc = "Ask this agent anything for help."
        
        server = FastMCP(name)

        async def ask_agent(message:str):
            return agent.input(message).start()
        server.tool(
            name=name,
            description=desc,
        )(ask_agent)
        
        server.run()
    
    def serve_workflow(self, workflow: Workflow, *, name:str=None, desc:str=None):
        if name is None or name == "":
            name = "Agently Workflow"
        if desc is None or desc == "":
            desc = "A helpful workflow to solve any problem you submit."
        
        server = FastMCP(name)

        async def submit_to_workflow(message:str, initial_storage:dict=None):
            result = workflow.start(message, storage=initial_storage)
            return result["default"]
        submit_to_workflow.__doc__ = desc
        server.tool()(submit_to_workflow)

        server.run()
    
    def serve(self, target: Union[Agent, Workflow], *, name:str=None, desc:str=None):
        if isinstance(target, Agent):
            self.serve_agent(target, name=name, desc=desc)
        elif isinstance(target, Workflow):
            self.serve_workflow(target, name=name, desc=desc)
        else:
            raise TypeError(f"[Agently FastServer] Can not support target '{ target }'. Server target must be an Agently Agent or a Agently Workflow.")


import os
from .utils import RequestABC, to_prompt_structure, to_instruction, to_json_desc, replace_placeholder_keyword, format_request_messages
from Agently.utils import RuntimeCtxNamespace

class Qianfan(RequestABC):
    def __init__(self, request):
        try:
            import qianfan
        except:
            raise ImportError("[Agently Request] Can not find package 'qianfan', please use `pip install qianfan` to install.")
        self.qianfan = qianfan
        self.request = request
        self.request_type = self.request.request_runtime_ctx.get("request_type", "chat")
        if self.request_type == None:
            self.request_type = "chat"
        self.model_name = "Qianfan"
        self.model_settings = RuntimeCtxNamespace(f"model.{self.model_name}", self.request.settings)
        if not self.model_settings.get_trace_back("message_rules.no_multi_system_messages"):
            self.model_settings.set("message_rules.no_multi_system_messages", True)
        if not self.model_settings.get_trace_back("message_rules.strict_orders"):
            self.model_settings.set("message_rules.strict_orders", True)
        if not self.model_settings.get_trace_back("message_rules.no_multi_type_messages"):
            self.model_settings.set("message_rules.no_multi_type_messages", True) 
    
    def construct_request_messages(self):
        #init request messages
        request_messages = []
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get_trace_back("prompt.general")
        if general_instruction_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[重要指导说明]\n{ to_instruction(general_instruction_data) }" }] })
        # - role
        role_data = self.request.request_runtime_ctx.get_trace_back("prompt.role")
        if role_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[角色及行为设定]\n{ to_instruction(role_data) }" }] })
        # - user info
        user_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.user_info")
        if user_info_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[用户信息]\n{ to_instruction(user_info_data) }" }] })
        # - abstract
        headline_data = self.request.request_runtime_ctx.get_trace_back("prompt.abstract")
        if headline_data:
            request_messages.append({ "role": "assistant", "content": [{"type": "text", "text": f"[主题及摘要]\n{ to_instruction(headline_data) }" }] })
        # - chat history
        chat_history_data = self.request.request_runtime_ctx.get_trace_back("prompt.chat_history")
        if chat_history_data:
            request_messages.extend(chat_history_data)
        # - request message (prompt)
        prompt_input_data = replace_placeholder_keyword("input", "输入", self.request.request_runtime_ctx.get_trace_back("prompt.input"))
        prompt_info_data = replace_placeholder_keyword("info", "补充信息", self.request.request_runtime_ctx.get_trace_back("prompt.info"))
        prompt_instruct_data = replace_placeholder_keyword("instruct", "处理规则", self.request.request_runtime_ctx.get_trace_back("prompt.instruct"))
        prompt_output_data = replace_placeholder_keyword("output", "输入", self.request.request_runtime_ctx.get_trace_back("prompt.output"))
        # --- only input
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception("[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        prompt_text = ""
        if prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            prompt_text = to_instruction(prompt_input_data)
        # --- construct prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[输入]"] = to_instruction(prompt_input_data)
            if prompt_info_data:
                prompt_dict["[补充信息]"] = str(prompt_info_data)
            if prompt_instruct_data:
                prompt_dict["[处理规则]"] = to_instruction(prompt_instruct_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[输出要求]"] = {
                        "TYPE": "可被解析的JSON字符串",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[输出要求]"] = str(prompt_output_data)
            prompt_text = to_prompt_structure(prompt_dict, end="[输出]:\n")
        request_messages.append({ "role": "user", "content": [{"type": "text", "text": prompt_text }] })
        return request_messages

    def construct_completion_prompt(self):
        # - init prompt
        completion_prompt = ""
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get_trace_back("prompt.general")
        if general_instruction_data:
            completion_prompt += f"[重要指导说明]\n{ to_instruction(general_instruction_data) }\n"
        # - role
        role_data = self.request.request_runtime_ctx.get_trace_back("prompt.role")
        if role_data:
            completion_prompt += f"[角色及行为设定]\n{ to_instruction(role_data) }\n"
        # - user info
        user_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.user_info")
        if user_info_data:
            completion_prompt += f"[用户信息]\n{ to_instruction(user_info_data) }\n"
        # - abstract
        headline_data = self.request.request_runtime_ctx.get_trace_back("prompt.abstract")
        if headline_data:
            completion_prompt += f"[主题及摘要]\n{ to_instruction(headline_data) }\n"
        # - main prompt
        chat_history_data = self.request.request_runtime_ctx.get_trace_back("prompt.chat_history")
        prompt_input_data = self.request.request_runtime_ctx.get_trace_back("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get_trace_back("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get_trace_back("prompt.output")
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data and not chat_history_data:
            raise Exception("[Request] Missing 'prompt.chat_history', 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        # --- only input
        if prompt_input_data and not chat_history_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            if completion_prompt == "":
                completion_prompt += to_instruction(prompt_input_data)
            else:
                completion_prompt += f"[输出]\n{ to_instruction(prompt_input_data) }"
        # --- construct prompt
        else:
            # - with output format
            if isinstance(prompt_output_data, (dict, list, set)):
                prompt_dict = {}
                if chat_history_data:
                    prompt_dict["[历史会话记录]"] = ""
                    for message in chat_history_data:
                        prompt_dict["[历史会话记录]"] += f"{ message['role'] }: { message['content'] }\n"
                if prompt_input_data:
                    prompt_dict["[输入]"] = to_instruction(prompt_input_data)
                if prompt_info_data:
                    prompt_dict["[补充信息]"] = str(prompt_info_data)
                if prompt_instruct_data:
                    prompt_dict["[处理规则]"] = to_instruction(prompt_instruct_data)
                if prompt_output_data:
                    prompt_dict["[输出要求]"] = {
                        "TYPE": "可被解析的JSON字符串",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                completion_prompt += to_prompt_structure(
                    prompt_dict,
                    end="[输出]:\nassistant: "
                        if chat_history_data and len(chat_history_data) > 0
                        else "[输出]:\n"
                )
            # without output format
            else:
                if prompt_info_data:
                    completion_prompt += f"[补充信息]\n{ str(prompt_info_data) }\n"
                if prompt_instruct_data:
                    completion_prompt += f"[处理规则]\n{ to_instruction(prompt_instruct_data) }\n"
                if prompt_output_data:
                    completion_prompt += f"[输出要求]\n{ str(prompt_output_data) }\n"
                completion_prompt += "[输出]\n"
                # chat completion
                if chat_history_data:
                    for message in chat_history_data:
                        completion_prompt += f"{ message['role'] }: { message['content'] }\n"
                    if prompt_input_data:
                        completion_prompt += f"user: { to_instruction(prompt_input_data) }\n"
                    completion_prompt += "assistant: "
                # text completion
                else:
                    if prompt_input_data:
                        completion_prompt += prompt_input_data
        return completion_prompt

    def generate_request_data(self):
        auth = self.model_settings.get_trace_back("auth", {})
        if "access_key" not in auth or "secret_key" not in auth:
            raise Exception(f"[Request] Qianfan require 'access_key' and 'secret_key'. Use .set_model_auth({{ 'access_key': <YOUR-ACCESS-KEY> }} and {{ 'secret_key': <YOUR-SECRET-KEY> }}) to set.")
        os.environ["QIANFAN_ACCESS_KEY"] = auth["access_key"]
        os.environ["QIANFAN_SECRET_KEY"] = auth["secret_key"]
        options = self.model_settings.get_trace_back("options", {})
        options = options if isinstance(options, dict) else {}
        if self.request_type == "chat":
            if "model" not in options:
                options["model"] = "ERNIE-Speed-8K"
            messages = format_request_messages(self.construct_request_messages(), self.model_settings)
            request_messages = []
            #system_prompt = ""
            for message in messages:
                if message["role"] == "system":
                    message["role"] = "user"
                    request_messages.append(message)
                    #system_prompt += f"{ message['content'] }\n"
                else:
                    request_messages.append(message)
            request_data = {
                "messages": format_request_messages(request_messages, self.model_settings),
                "stream": True,
                **options,
            }
            #if system_prompt != "" and self.request.settings.get_trace_back("retry_count", 0) > 0:
            #    request_data.update({ "system": system_prompt })
        elif self.request_type in ["completion", "completions"]:
            if "model" not in options:
                options["model"] = "Yi-34B-Chat"
            prompt = self.construct_completion_prompt()
            request_data = {
                "prompt": prompt,
                "stream": True,
                **options
            }
        elif self.request_type in ["embedding", "embeddings"]:
            if "model" not in options:
                options["model"] = "Embedding-V1"
            content_input = (
                self.request.request_runtime_ctx.get("prompt.input")
                or self.request.request_runtime_ctx.get("prompt.texts")
            )
            if not isinstance(content_input, list):
                content_input = [content_input]
            request_data = {
                "texts": content_input,
                **options,
            }
        elif self.request_type in ["image", "images"]:
            if "model" not in options:
                options["model"] = "Stable-Diffusion-XL"
            prompt = (
                self.request.request_runtime_ctx.get("prompt.input")
                or self.request.request_runtime_ctx.get("prompt.prompt")
            )
            request_data = {
                "prompt": prompt,
                **options
            }
        elif self.request_type in ["rerank", "reranker"]:
            if "model" not in options:
                options["model"] = "bce-reranker-base_v1"
            input_data = self.request.request_runtime_ctx.get("prompt.input")
            query = (
                input_data["query"]
                if "query" in input_data
                else self.request.request_runtime_ctx.get("prompt.query")
            )
            doucments = (
                input_data["documents"]
                if "documents" in input_data
                else self.request.request_runtime_ctx.get("prompt.documents")
            )
            request_data = {
                "query": query,
                "documents": doucments,
                **options,
            }
        return request_data
    
    async def request_model(self, request_data: dict):
        if self.request_type == "chat":
            return await self.qianfan.ChatCompletion().ado(**request_data)
        elif self.request_type in ["completion", "completions"]:
            return await self.qianfan.Completion().ado(**request_data)
        elif self.request_type in ["embedding", "embeddings"]:
            return await self.qianfan.Embedding().ado(**request_data)
        elif self.request_type in ["image", "images"]:
            return await self.qianfan.Text2Image().ado(**request_data)
        elif self.request_type in ["rerank", "reranker"]:
            return await self.qianfan.resources.Reranker().ado(**request_data)
    
    async def broadcast_response(self, response_generator):
        if self.request_type == "chat" or self.request_type in ["completion", "completions"]:
            origin_buffer = {}
            buffer = ""
            async for part in response_generator:
                buffer += part["body"]["result"]
                origin_buffer.update(part["body"])
                yield ({ "event": "response:delta_origin", "data": part["body"] })
                yield ({ "event": "response:delta", "data": part["body"]["result"] })
            origin_buffer["result"] = buffer
            yield ({ "event": "response:done_origin", "data": origin_buffer })
            yield ({ "event": "response:done", "data": buffer })
        elif self.request_type in ["embedding", "embeddings"] or self.request_type in ["image", "images"]:
            yield({ "event": "response:done_origin", "data": response_generator["body"] })
            yield({ "event": "response:done", "data": response_generator["body"]["data"] })
        elif self.request_type in ["rerank", "reranker"]:
            yield({ "event": "response:done_origin", "data": response_generator.body })
            yield({ "event": "response:done", "data": response_generator.body["results"] })
    
    def export(self):
        return {
            "generate_request_data": self.generate_request_data,
            "request_model": self.request_model,
            "broadcast_response": self.broadcast_response,
        }


def export():
    return ("Qianfan", Qianfan)


from typing import TypeAlias, Dict, List, TypedDict, Literal


class ChunkDepData(TypedDict):
  handle: str
  data_slots: List[Dict]

class ChunkData(TypedDict):
  id: str
  loop_entry: bool
  type: Literal['Normal', 'Loop', 'Start', 'End']
  executor: callable
  deps: List[ChunkDepData]


import inspect
from typing import TYPE_CHECKING
from ..executors.generater.loop import use_loop_executor
from ..lib.constants import EXECUTOR_TYPE_LOOP
from .helper import exposed_interface

if TYPE_CHECKING:
    from . import SchemaChunk

class LoopAbility:
    """提供Loop API"""

    def __init__(self) -> None:
        self.shared_ns['loop'] = {}
        super().__init__()

    @exposed_interface(type='connect_command')
    def loop_with(self, sub_workflow) -> 'SchemaChunk':
        """遍历逐项处理，支持传入子 workflow/处理方法 作为处理逻辑"""
        is_function = inspect.isfunction(sub_workflow) or inspect.iscoroutinefunction(sub_workflow)
        # 这里是新的 chunk，需要走 Schema 的 create 方法挂载
        loop_chunk = self.create_rel_chunk(
            type=EXECUTOR_TYPE_LOOP,
            title="Loop",
            executor=use_loop_executor(sub_workflow),
            extra_info={
                # loop 的相关信息
                "loop_info": {
                    "type": 'function' if is_function else 'workflow',
                    "detail": sub_workflow.__name__ if is_function else sub_workflow.schema.compile()
                }
            }
        )
        return self._raw_connect_to(loop_chunk)



from .utils import ComponentABC
from Agently.utils import RuntimeCtxNamespace

class Status(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.status_runtime_ctx = RuntimeCtxNamespace("status", self.agent.agent_runtime_ctx)
        self.status_mapping_runtime_ctx = RuntimeCtxNamespace("status_mapping", self.agent.agent_runtime_ctx)
        self.status_storage = self.agent.global_storage.table("status")
        self.settings = RuntimeCtxNamespace("plugin_settings.agent_component.Status", self.agent.settings)
        self.global_status_namespace = None

        # Defer loading to avoid unnecessary actions in initialization
        if self.settings.get_trace_back("auto"):
            self.load()

    def set(self, key: str, value: str):
        self.status_runtime_ctx.set(key, value)
        return self.agent

    def use_global_status(self, namespace_name: str="default"):
        self.global_status_namespace = namespace_name
        return self.agent

    def append_mapping(self, status_key: str, status_value: str, alias_name: str, *args, **kwargs):
        self.status_mapping_runtime_ctx.append(
            f"{status_key}.{status_value}",
            {
                "alias_name": alias_name,
                "args": args,
                "kwargs": kwargs,
            }
        )
        return self.agent

    def save(self):
        status = self.status_runtime_ctx.get()
        self.status_storage.set(status).save()
        return self.agent

    def load(self):
        status = self.status_storage.get()
        self.status_runtime_ctx.update(status)
        return self.agent

    def _apply_mappings(self, mappings_dict, status_key, status_value):
        """Helper function to apply mappings to the agent."""
        if status_key in mappings_dict and status_value in mappings_dict[status_key]:
            for mapping in mappings_dict[status_key][status_value]:
                alias_func = getattr(self.agent, mapping["alias_name"], None)
                if alias_func:
                    alias_func(*mapping["args"], **mapping["kwargs"])

    def _prefix(self):
        agent_status = self.status_runtime_ctx.get()
        if not agent_status:
            return None

        # Load global mappings if the namespace is set
        global_mappings = {}
        if self.global_status_namespace:
            global_mappings = self.agent.global_storage.table(f"status_mapping.{self.global_status_namespace}").get() or {}

        agent_mappings = self.status_mapping_runtime_ctx.get() or {}

        for status_key, status_value in agent_status.items():
            self._apply_mappings(global_mappings, status_key, status_value)
            self._apply_mappings(agent_mappings, status_key, status_value)

        return None

    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": None,
            "alias": {
                "set_status": {"func": self.set},
                "save_status": {"func": self.save},
                "load_status": {"func": self.load},
                "use_global_status": {"func": self.use_global_status},
                "append_status_mapping": {"func": self.append_mapping},
            },
        }

def export():
    return ("Status", Status)



class StageFunction:
    def __init__(self, stage, func, *args, **kwargs):
        self._stage = stage
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self._on_success_handler = None
        self._on_error_handler = None
    
    def set(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        return self
    
    def args(self, *args):
        self._args = args
        return self
    
    def kwargs(self, **kwargs):
        self._kwargs = kwargs
        return self
    
    def on_success(self, on_success_handler):
        self._on_success_handler = on_success_handler
        return self

    def on_error(self, on_error_handler):
        self._on_error_handler = on_error_handler
        return self

    def go(self):
        return self._stage.go(
            self._func,
            *self._args,
            on_success=self._on_success_handler,
            on_error=self._on_error_handler,
            **self._kwargs
        )
    
    def get(self):
        return self._stage.get(
            self._func,
            *self._args,
            on_success=self._on_success_handler,
            on_error=self._on_error_handler,
            **self._kwargs
        )

class StageFunctionMixin:        
    def func(self, func):
        return StageFunction(self, func)


import asyncio
from itertools import combinations
from .utils import ComponentABC
from Agently.utils import RuntimeCtx, RuntimeCtxNamespace

class EventListener(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.agent_listeners = RuntimeCtxNamespace("event_listeners", self.agent.agent_runtime_ctx)
        self.listeners = RuntimeCtxNamespace("event_listeners", RuntimeCtx(parent=self.agent.agent_runtime_ctx))
        self.listeners.set({})
        self.async_tasks = []

    def add(self, event:str, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        event = event.replace(".", "->")
        if event == "realtime":
            self.agent.settings.set("use_instant", True)
            event = "instant"
        if event.startswith("realtime:"):
            self.agent.settings.set("use_instant", True)
            event = "instant:" + event[9:]
        if event == "instant" or event.startswith("instant:"):
             self.agent.settings.set("use_instant", True)
        if event.startswith("instant:"):
            event_data = event.replace(" ", "").split(":")
            keys = event_data[1].replace("->", ".").split("&")
            key_indexes_list = []
            for key_str in keys:
                if isinstance(key_str, str):
                    if "?" in key_str:
                        key, indexes_str = key_str.split("?")
                        index_list = indexes_str.split(",")
                        if index_list == [""]:
                            index_list = []
                    else:
                        key = key_str
                        index_list = []
                    indexes = []
                    for index in index_list:
                        if index in ("_", "*"):
                            indexes.append(-1)
                        else:
                            indexes.append(int(index))
                    key_indexes_list.append((key, indexes))
            async def instant_hook_handler(data):
                indexes = data["indexes"]
                if (data["key"], indexes) in key_indexes_list or (data["key"], []) in key_indexes_list:
                    if asyncio.iscoroutinefunction(listener):
                        await listener(data)
                    else:
                        listener(data)
                indexes_len = len(indexes)
                for r in range(1, indexes_len + 1):
                    for indices in combinations(range(indexes_len), r):
                        possible_indexes = indexes[:]
                        for i in indices:
                            possible_indexes[i] = -1
                        if (data["key"], possible_indexes) in key_indexes_list:
                            if asyncio.iscoroutinefunction(listener):
                                await listener(data)
                            else:
                                listener(data)
            if event not in (self.listeners.get(trace_back=False) or {}):
                self.listeners.update(event, [])
            self.listeners.append("instant", { "listener": instant_hook_handler, "is_await": is_await })
        else:
            if is_agent_event:
                if event not in (self.agent_listeners.get(trace_back=False) or {}):
                    self.agent_listeners.update(event, [])
                self.agent_listeners.append(event, { "listener": listener, "is_await": is_await })
                return self.agent
            else:
                if event not in (self.listeners.get(trace_back=False) or {}):
                    self.listeners.update(event, [])
                self.listeners.append(event, { "listener": listener, "is_await": is_await })
                return self.agent

    def on_delta(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.add("response:delta", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent

    def on_done(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.add("response:done", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent

    def on_finally(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.add("response:finally", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent

    def on_instant(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.add("instant", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent
    
    def on_realtime(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.agent.settings.set("use_instant", True)
        self.add("instant", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent
    
    async def call_event_listeners(self, event: str, data: any):
        event = event.replace(".", "->")
        listeners = self.listeners.get_trace_back() or {}
        if "*" in listeners:
            for listener_info in listeners["*"]:
                if asyncio.iscoroutinefunction(listener_info["listener"]):
                    if listener_info["is_await"]:
                        await listener_info["listener"](event , data)
                    else:
                        self.async_tasks.append(asyncio.create_task(listener_info["listener"](event, data)))
                else:
                    listener_info["listener"](event, data)
        if event in listeners:
            for listener_info in listeners[event]:
                if asyncio.iscoroutinefunction(listener_info["listener"]):
                    if listener_info["is_await"]:
                        await listener_info["listener"](data)
                    else:
                        self.async_tasks.append(asyncio.create_task(listener_info["listener"](data)))
                else:
                    listener_info["listener"](data)
        return self.agent

    async def _suffix(self, event:str, data: any):
        await self.call_event_listeners(event, data)
        if event == "response:finally":
            for async_task in self.async_tasks:
                await async_task
            self.async_tasks = []
            self.listeners.set({})

    def export(self):
        return {
            "prefix": None,
            "suffix": self._suffix,
            "alias": {
                "add_event_listener": { "func": self.add },
                "on_delta": { "func": self.on_delta },
                "on_done": { "func": self.on_done },
                "on_finally": { "func": self.on_finally },
                "on_instant": { "func": self.on_instant },
                "call_event_listeners": { "func": self.call_event_listeners },
            },
        }

def export():
    return ("EventListener", EventListener)


from typing import Union, Literal
from Agently.Agent import Agent
from Agently.Workflow import Workflow

class FastServer:
    #def __init__(self, type: Literal["mcp", "websocket", "sse", "http"]):
    def __init__(self, type: Literal["mcp"]):
        if type == "mcp":
            from .builtins import MCPServerHandler as ServerHandler
        elif type == "websocket":
            from .builtins import WebSocketServerHandler as ServerHandler
        elif type == "sse":
            from .builtins import SSEServerHandler as ServerHandler
        elif type == "http":
            from .builtins import HTTPServerHandler as ServerHandler
        else:
            raise TypeError(f"[Agently FastServer] Can not support type '{ type }.'")
        self._server_handler = ServerHandler()
        self.serve = self._server_handler.serve


#ABC = Abstract Base Class
from abc import ABC, abstractmethod

class ComponentABC(ABC):
    def __init__(self, agent):
        self.agent = agent

    @abstractmethod
    def export(self):
        return {
            "prefix": callable or [callable],#()->(request_namespace: str, update_dict: dict)
            "suffix": callable,#(event: str, data: any)->None
            "alias": {
                alias_name: {
                    "func": callable,
                    "return_value": bool,# optional, default=False
                },
                # ...
            },
        }




from .utils import FacilityABC
from Agently.utils import RuntimeCtx, RuntimeCtxNamespace


class Embedding(FacilityABC):
    def __init__(self, *, storage: object, plugin_manager: object, settings: object):
        self.plugin_manager = plugin_manager
        self.settings = settings

    def OpenAI(self, text: str, options: dict = {}):
        from openai import OpenAI
        import httpx

        settings = RuntimeCtxNamespace("embedding.OpenAI", self.settings)
        # Prepare Client Params
        client_params = {}
        base_url = settings.get_trace_back("url")
        if base_url:
            client_params.update({"base_url": base_url})
        proxy = self.settings.get_trace_back("proxy")
        if proxy:
            client_params.update({"http_client": httpx.Client(proxies=proxy)})
        api_key = settings.get_trace_back("auth.api_key")
        if api_key:
            client_params.update({"api_key": api_key})
        else:
            raise Exception(
                "[Embedding] OpenAI require api_key. use Agently.facility.set_settings('embedding.OpenAI.auth', { 'api_key': '<Your-API-Key>' }) to set it."
            )
        request_options = settings.get_trace_back("options")
        request_options.update(options)
        if "model" not in request_options:
            request_options["model"] = "text-embedding-ada-002"
        # Create Client
        client = OpenAI(**client_params)
        # Request
        result = client.embeddings.create(
            input=text.replace("\n", " "), **request_options
        )
        if result.data:
            return result.data[0].embedding
        else:
            raise Exception(f"[Embedding] OpenAI Error: { dict(result) }")

    def ERNIE(self, text_list: (str, list), options: dict = {}):
        import erniebot

        settings = RuntimeCtxNamespace("embedding.ERNIE", self.settings)
        erniebot.api_type = "aistudio"
        access_token = settings.get_trace_back("auth.aistudio")
        if access_token:
            erniebot.access_token = access_token
        else:
            raise Exception(
                "[Embedding] ERNIE require aistudio access token. use Agently.facility.set_settings('embedding.ERNIE.auth', { 'aistudio': '<Your-Access-Token>' }) to set it."
            )
        if "model" not in options:
            options["model"] = "ernie-text-embedding"
        if not isinstance(text_list, list):
            text_list = [text_list]
        result = erniebot.Embedding.create(input=text_list, **options)
        result = dict(result)
        if result["rcode"] == 200:
            return result["rbody"]["data"]
        else:
            raise Exception(f"[Embedding] ERNIE Error: { result }")

    def ZhipuAI(self, text: str, options: dict = {}):
        import zhipuai

        settings = RuntimeCtxNamespace("embedding.ZhipuAI", self.settings)
        api_key = settings.get_trace_back("auth.api_key")
        if api_key:
            zhipuai.api_key = api_key
        else:
            raise Exception(
                "[Embedding] ZhipuAI require api_key. use Agently.facility.set_settings('embedding.ZhipuAI.auth', { 'api_key': '<Your-API-Key>' }) to set it."
            )
        if "model" not in options:
            options["model"] = "text_embedding"
        result = zhipuai.model_api.invoke(prompt=text, **options)
        if "data" in result:
            return result["data"]["embedding"]
        else:
            raise Exception(f"[Embedding] ZhipuAI Error: { result }")

    def Google(self, text: str, options: dict = {}):
        import json
        import httpx

        settings = RuntimeCtxNamespace("embedding.Google", self.settings)
        api_key = settings.get_trace_back("auth.api_key")
        if api_key == None:
            raise Exception(
                "[Embedding] Google require api_key. use Agently.facility.set_settings('embedding.google.auth', { 'api_key': '<Your-API-Key>' }) to set it."
            )
        proxy = self.settings.get_trace_back("proxy")
        model = "embedding-001"
        if "model" in options:
            model = options["model"]
            del options["model"]
        request_params = {
            "data": json.dumps({"content": {"parts": [{"text": text}]}, **options}),
            "timeout": None,
        }
        result = httpx.post(
            f"https://generativelanguage.googleapis.com/v1/models/{ model }:embedContent?key={ api_key }",
            **request_params,
        )
        if result.status_code:
            return result.json()["embedding"]["values"]
        else:
            raise Exception(f"[Embedding] Google Error: { dict(result) }")


def export():
    return ("embedding", Embedding)



import uuid
from typing import TYPE_CHECKING
from ..lib.constants import EXECUTOR_TYPE_CONDITION
from .helper import exposed_interface

if TYPE_CHECKING:
    from . import SchemaChunk

class ConditionAbility:
    """提供条件判断 API"""

    def __init__(self, **chunk_desc):
        self.shared_chain_ns['condition_ns'] = {
            'condition_chunk_path': chunk_desc.get('condition_chunk_path', [])
        }
        super().__init__()

    @exposed_interface(type='connect_command')
    def if_condition(self, condition: callable = None) -> 'SchemaChunk':
        """条件判断"""
        condition_signal_id = str(uuid.uuid4())
        # 创建新的 condition chunk，需要走 Schema 的 create 方法挂载
        condition_chunk = self.create_rel_chunk(
            type=EXECUTOR_TYPE_CONDITION,
            title="Condition",
            extra_info={
                # condition 的相关信息
                "condition_info": {
                    # 常规条件
                    "conditions": [{
                        # 命中时的结果符号
                        "condition_signal": condition_signal_id,
                        "condition": condition
                    }],
                    "else_condition": {
                        # 命中时的结果符号
                        "condition_signal": str(uuid.uuid4())
                    }
                }
            }
        )
        new_path_item = {
            'root': self,
            'condition': condition_chunk,
            # 关联的句柄(本质也是Chunk实例)
            'handles': []
        }
        # 以 condition_chunk 为起点，追加上条件路径信息（下游条件逻辑都可基于此判断处理）
        condition_chunk.shared_chain_ns['condition_ns']['condition_chunk_path'].append(new_path_item)
        # 把上游chunk连接到主判断节点（仅 if 时操作一次即可，后续皆为连接到condition chunk）
        connected_chunk_handle = self._raw_connect_to(condition_chunk)
        # 新连接的第一个 chunk，也需要补充本次的条件路径
        connected_chunk_handle.shared_chain_ns['condition_ns']['condition_chunk_path'].append(new_path_item)
        # 设置条件参数和辅助信息
        connected_chunk_handle.set_connect_condition(
            condition=lambda runtime_signal, storage: runtime_signal == condition_signal_id,
            condition_detail={
                # condition id
                "id": f"cid-{str(uuid.uuid4())}",
                "type": "if"
            }
        )
        return connected_chunk_handle

    @exposed_interface(type='connect_command')
    def elif_condition(self, condition: callable = None) -> 'SchemaChunk':
        """elif 条件判断"""
        # Step1. 从当前 chunk 中，读取所属的条件 chunk id 信息
        condition_path = self.shared_chain_ns['condition_ns']['condition_chunk_path']
        condition_path_item = condition_path[-1] if condition_path else None
        if not condition_path_item:
            raise ValueError(
                f"The `elif_condition` method must be called after the `if_condition` method.")

        condition_chunk: 'SchemaChunk' = condition_path_item.get('condition')
        # Step2. 基于 condition_chunk 做进一步的连接和更新操作
        # 2.1 生成 elif_fn
        condition_signal_id = str(uuid.uuid4())
        condition_info = (condition_chunk.chunk.get(
            'extra_info') or {}).get('condition_info') or {}
        # 追加条件信号
        condition_info.get('conditions', []).append({
            # 命中时的结果符号
            "condition_signal": condition_signal_id,
            "condition": condition
        })

        # 2.2 进行条件连接
        # 设置条件链接信息
        condition_chunk.set_connect_condition(
            condition=lambda runtime_signal, storage: runtime_signal == condition_signal_id,
            condition_detail={
                "id": (condition_chunk.chunk.get('connect_condition_detail') or {}).get('id') or f"cid-{str(uuid.uuid4())}",
                "type": "elif"
            }
        )
        shadow_root_chunk = condition_chunk.copy_shadow_chunk()
        # 将 elif 的上一个端点存入 handles 点中（后续结束时要统一处理）
        shadow_root_chunk.save_condition_end_handles(self)
        return shadow_root_chunk

    @exposed_interface(type='connect_command')
    def else_condition(self) -> 'SchemaChunk':
        """
        按当前条件的反条件连接（特别注意，else_condition 作用的 chunk 为往前追溯的最近一个 if_condition 作用的 chunk，两是同一个，成套存在的）
        """
        # Step1. 从当前 chunk 中，读取所属的条件 chunk id 信息
        condition_path = self.shared_chain_ns['condition_ns']['condition_chunk_path']
        condition_path_item = condition_path[-1] if condition_path else None
        if not condition_path_item:
            raise ValueError(
                f"The `else_condition` method must be called after the `if_condition` method.")

        condition_chunk: 'SchemaChunk' = condition_path_item.get('condition')
        # Step2. 基于 condition_chunk 做进一步的连接和更新操作
        # 2.1 生成 else_fn
        condition_info = (condition_chunk.chunk.get(
            'extra_info') or {}).get('condition_info') or {}
        else_signal = (condition_info.get('else_condition') or {}).get(
            'condition_signal') or str(uuid.uuid4())

        # 2.2 进行条件连接
        # 在 shadow_root_chunk 上操作，支撑参数剔除当前所在条件 chunk 后继续保留
        shadow_root_chunk = condition_chunk.copy_shadow_chunk()
        # 设置条件链接信息
        shadow_root_chunk.set_connect_condition(
            condition=lambda runtime_signal, storage: runtime_signal == else_signal,
            condition_detail={
                "id": (condition_chunk.chunk.get('connect_condition_detail') or {}).get('id') or f"cid-{str(uuid.uuid4())}",
                "type": "else"
            }
        )
        # 将 else 的上一个端点存入 handles 点中（后续结束时要统一处理）
        shadow_root_chunk.save_condition_end_handles(self)
        return shadow_root_chunk

    @exposed_interface(type='connect_command')
    def end_condition(self) -> 'SchemaChunk':
        """终止当前所在的if，回退到当次 if 前的 chunk"""
        # Step1. 从当前 chunk 中，读取所属的条件 chunk id 信息
        condition_path = self.shared_chain_ns['condition_ns']['condition_chunk_path']
        condition_path_item = condition_path[-1] if condition_path else None
        if not condition_path_item:
            raise ValueError(
                f"The `end_condition` method must be called after the `if_condition` method.")
        # Step2. 将 end 的上一个端点存入 handles 点中（后续结束时要统一处理）
        shadow_chunk = self.copy_shadow_chunk()
        shadow_chunk.save_condition_end_handles(self)
        # Step3. 汇总所有 handles，统一连接到下一个节点（未来）
        def future_commands(chunk: 'SchemaChunk', type, name, res):
            # 针对 command 类型，执行future
            if type == 'connect_command':
                condition_end_handles = chunk.get_condition_end_handles()
                # 将本次 condition的所有末级端点批量连接到 end_condition 之后的端点上
                if condition_end_handles:
                    for end_handle in condition_end_handles:
                        # 跳过自身，只处理其它端点
                        if end_handle.chunk['id'] == shadow_chunk.chunk['id']:
                            continue
                        # 调用其它端点的处理逻辑
                        end_handle._raw_connect_to(res.copy_shadow_chunk())

            # 针对其它类型报错处理
            else:
                raise TypeError("After end_condition, the method cannot be called directly.")

        shadow_chunk._post_command_interceptors.append(future_commands)
        return shadow_chunk

    def set_connect_condition(self, condition: callable = None, condition_detail: dict = None):
        """设置连接的条件"""
        self.chunk['connect_condition'] = condition
        self.chunk['connect_condition_detail'] = condition_detail
        return self
    
    def get_condition_end_handles(self):
        """获取当前条件的结束的句柄 chunk（注意，此句柄是挂载在 condition_chunk 上的）"""
        # 找到 condtion_chunk 的handles
        condition_path = self.shared_chain_ns['condition_ns']['condition_chunk_path']
        if not condition_path:
            return []
        condtion_chunk = self.shared_chain_ns['condition_ns']['condition_chunk_path'][-1]['condition']
        current_handles = condtion_chunk.shared_chain_ns[
            'condition_ns']['condition_chunk_path'][-1]['handles']
        return current_handles

    def save_condition_end_handles(self, handle_chunk: 'SchemaChunk'):
        """保存 condition 的各连接端点句柄"""
        # 找到 condtion_chunk 的handles
        condtion_chunk = self.shared_chain_ns['condition_ns']['condition_chunk_path'][-1]['condition']
        current_handles = condtion_chunk.shared_chain_ns['condition_ns']['condition_chunk_path'][-1]['handles']
        # 如果已存在，则忽略
        for handle in current_handles:
            if handle.chunk['id'] == handle_chunk.chunk['id']:
                return self
        current_handles.append(handle_chunk)
        return self



import uuid
from .Stage import Stage

class MessageCenter:
    def __init__(self, private_max_workers=None, max_concurrent_tasks=None, on_error=None):
        self._stage = Stage(
            private_max_workers=private_max_workers,
            max_concurrent_tasks=max_concurrent_tasks,
            on_error=on_error,
            is_daemon=True
        )
        self._consumers = {}
    
    def register_consumer(self, handler):
        consumer_id = uuid.uuid4()
        self._consumers.update({ consumer_id: handler })
        return consumer_id
    
    def remove_consumer(self, consumer_id):
        del self._consumers[consumer_id]
    
    def put(self, data):
        for _, consumer_handler in self._consumers.items():
            self._stage.go(consumer_handler, data)
    
    def close(self):
        self._stage.close()


import queue
import asyncio
import threading

class StageHybridGenerator:
    def __init__(self, stage, task, on_success=None, on_error=None, lazy=None, async_gen_interval=0.1):
        self._stage = stage
        self._stage._responses.add(self)
        self._loop = stage._loop
        self._on_success = on_success
        self._on_error = on_error
        self._task = task
        self._result = []
        self._error = None
        self._result_queue = queue.Queue()
        self._result_ready = threading.Event()
        self._completed = False
        self._final_result = None
        self._is_lazy = lazy
        self._async_gen_interval = async_gen_interval
        if not self._is_lazy:
            self._run_consume_async_gen(self._task)
    
    def _run_consume_async_gen(self, task):
        consume_result = asyncio.run_coroutine_threadsafe(self._consume_async_gen(task), self._loop)
        consume_result.add_done_callback(self._on_consume_async_gen_done)
    
    def _on_consume_async_gen_done(self, future):
        future.result()
        if self._error is not None:
            def raise_error():
                raise self._error
            self._loop.call_soon_threadsafe(raise_error)
        if self._on_success:
            self._final_result = self._on_success(self._result)
        self._result_ready.set()
        self._stage._responses.discard(self)
    
    async def _consume_async_gen(self, task):
        try:
            async for item in task:
                self._result_queue.put(item)
                self._result.append(item)
            self._completed = True
        except Exception as e:
            if self._on_error:
                handled_result = self._on_error(e)
                self._result_queue.put(handled_result)
                self._result.append(handled_result)
            else:
                self._result_queue.put(e)
                self._result.append(e)
                self._error = e
        finally:
            self._result_queue.put(StopIteration)
    
    async def __aiter__(self):
        if self._is_lazy:
            self._run_consume_async_gen(self._task)
        while True:
            try:
                item = self._result_queue.get_nowait()
                if item is StopIteration:
                    break
                yield item
            except queue.Empty:
                await asyncio.sleep(self._async_gen_interval)

    def __iter__(self):
        if self._is_lazy:
            self._run_consume_async_gen(self._task)
        while True:
            item = self._result_queue.get()
            if item is StopIteration:
                break
            yield item
    
    def get(self):
        self._result_ready.wait()
        return self._result

    def get_final(self):
        self._result_ready.wait()
        if self._final_result:
            return self._final_result
        else:
            return self._result


from typing import Dict, Union
from contextlib import AsyncExitStack
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client

class MCPClient:
    def __init__(self, server_params: Union[Dict, StdioServerParameters]=None):
        self._server_params = (
            server_params
            if isinstance(server_params, StdioServerParameters) or server_params is None
            else StdioServerParameters(**server_params)
        )
        self._exit_stack = AsyncExitStack()
        self._read = None
        self._write = None
        self._session = None
    
    async def __aenter__(self):
        await self.connect()
        return self.session
    
    async def __aexit__(self, _, __, ___):
        await self.close()

    async def connect(self, server_params: StdioServerParameters=None):
        server_params = server_params or self._server_params
        if not server_params:
            raise RuntimeError("MCP server parameters are required.")
        self._read, self._write = await self._exit_stack.enter_async_context(
            stdio_client(server_params)
        )
        self._session: ClientSession = await self._exit_stack.enter_async_context(ClientSession(self._read, self._write))
        await self._session.initialize()
    
    async def close(self):
        await self._exit_stack.aclose()

    @property
    def session(self):
        if not self._session:
            raise RuntimeError("Use .connect(<server_params>) to connect a MCP server first.")
        return self._session
    
    def _transform_schema_to_type(self, schema):
        if isinstance(schema, dict):
            # Handle `anyOf` / `oneOf`
            if 'anyOf' in schema or 'oneOf' in schema:
                key = 'anyOf' if 'anyOf' in schema else 'oneOf'
                return " | ".join(self._transform_schema_to_type(sub) for sub in schema[key])

            # Handle enums
            if 'enum' in schema:
                return " | ".join(f'"{v}"' for v in schema['enum'])

            # Handle const
            if 'const' in schema:
                val = schema['const']
                return f"{repr(val)}"

            # Handle type
            t = schema.get('type')
            if t == 'string':
                return "string"
            elif t == 'number' or t == 'integer':
                return "number"
            elif t == 'boolean':
                return "boolean"
            elif t == 'null':
                return "null"
            elif t == 'array':
                items = schema.get('items', {'type': 'any'})
                return f"Array<{self._transform_schema_to_type(items)}>"
            elif t == 'object':
                props = schema.get('properties', {})
                required = schema.get('required', [])
                parts = []
                for key, val in props.items():
                    optional = '?' if key not in required else ''
                    parts.append(f"{key}{optional}: {self._transform_schema_to_type(val)}")
                return f"{{{', '.join(parts)}}}"
            elif t == 'any':
                return "any"

        return "any"

    async def get_tools_info(self):
        tools = await self.session.list_tools()
        tools_info = []
        for tool in tools.tools:
            tools_info.append({
                "tool_name": tool.name,
                "desc": tool.description,
                "kwargs": {}
            })
            for key, value in tool.inputSchema["properties"].items():
                type = self._transform_schema_to_type(value)
                desc = (
                    value["description"]
                    if "description" in value
                    else (
                        value["desc"]
                        if "desc" in value
                        else None
                    )
                )
                tools_info[-1]["kwargs"].update({
                    key: (type, desc),
                })
        return tools_info
    
    async def call_tool(self, tool_name: str, kwargs: dict):
        response = await self.session.call_tool(tool_name, kwargs)
        return response.content[0].text


from itertools import combinations
from .utils import ComponentABC
from Agently.Stage import Stage, Tunnel
from contextlib import contextmanager

class ResponseGenerator(ComponentABC):
    def __init__(self, agent):
        super().__init__(agent)
        self._tunnel = Tunnel()

    @contextmanager
    def _create_start_stage(self):
        try:
            stage = Stage(is_daemon=True)
            yield stage.func(self.agent.start)
        finally:
            stage.close()

    def put_data_to_generator(self, event, data):
        if (event, data) != (None, None):
            self._tunnel.put((event, data))
        else:
            self._tunnel.put_stop()

    def get_complete_generator(self):
        with self._create_start_stage() as stage:
            stage.go()
            for item in self._tunnel.get():
                yield item
            self._tunnel = Tunnel()
    
    def get_instant_keys_generator(self, keys):
        if not isinstance(keys, list):
            if isinstance(keys, str):
                keys = keys.split("&")
            else:
                raise Exception("[Response Generator]", ".get_instant_keys_generator(<keys>) require a list or string input.\nKey format: <key string>?<indexes string split by ','>")
        key_indexes_list = []
        for key_str in keys:
            if isinstance(key_str, str):
                if "?" in key_str:
                    key, indexes_str = key_str.split("?")
                    index_list = indexes_str.split(",")
                    if index_list == [""]:
                        index_list = []
                else:
                    key = key_str
                    index_list = []
                indexes = []
                for index in index_list:
                    if index in ("_", "*"):
                        indexes.append(-1)
                    else:
                        indexes.append(int(index))
                key_indexes_list.append((key, indexes))
        self.agent.settings.set("use_instant", True)
        with self._create_start_stage() as stage:
            stage.go()
            for item in self._tunnel.get():
                if item[0] == "instant":
                    indexes = item[1]["indexes"]
                    if (item[1]["key"], indexes) in key_indexes_list or (item[1]["key"], []) in key_indexes_list:
                        yield item[1]
                        continue
                    indexes_len = len(indexes)
                    for r in range(1, indexes_len + 1):
                        for indices in combinations(range(indexes_len), r):
                            possible_indexes = indexes[:]
                            for i in indices:
                                possible_indexes[i] = -1
                            if (item[1]["key"], possible_indexes) in key_indexes_list:
                                yield item[1]
                                break
            self._tunnel = Tunnel()
    
    def get_instant_generator(self):
        self.agent.settings.set("use_instant", True)
        try:
            with self._create_start_stage() as stage:
                stage.go()
                for item in self._tunnel.get():
                    if item[0] == "instant":
                        yield item[1]
        finally:
            self._tunnel = Tunnel()
    
    def get_generator(self):
        try:
            with self._create_start_stage() as stage:
                stage.go()
                for item in self._tunnel.get():
                    if not item[0].endswith(("_origin")):
                        yield item
        finally:
            self._tunnel = Tunnel()
    
    def get_delta_generator(self):
        with self._create_start_stage() as stage:
            stage.go()
            for event, data in self._tunnel.get():
                if event == "response:delta":
                    yield data
            self._tunnel = Tunnel()
        
    async def _suffix(self, event, data):
        if event != "response:finally":
            self._tunnel.put((event, data))
        else:
            self._tunnel.put_stop()
    
    def export(self):
        return {
            "suffix": self._suffix,
            "alias": { 
                "put_data_to_generator": { "func": self.put_data_to_generator },
                "get_generator": { "func": self.get_generator, "return_value": True },
                "get_delta_generator": { "func": self.get_delta_generator, "return_value": True },
                "get_instant_generator": { "func": self.get_instant_generator, "return_value": True },
                "get_realtime_generator": { "func": self.get_instant_generator, "return_value": True },
                "get_instant_keys_generator": { "func": self.get_instant_keys_generator, "return_value": True },
                "get_complete_generator": { "func": self.get_complete_generator, "return_value": True },
            },
        }

def export():
    return ("ResponseGenerator", ResponseGenerator)


import copy
from typing import Any, Dict, List, Optional, Union


class NamespaceOps:
    def __init__(self, namespace_name: str, data_ops: 'DataOps', *, return_to: Optional[Any] = None):
        self.namespace_name = namespace_name
        self.data_ops = data_ops
        self.return_to = return_to if return_to else self

    def assign(self, input: Any, desc: Any = None) -> 'NamespaceOps':
        """智能分配值，根据输入类型自动选择合适的方法"""
        if desc is None:
            if isinstance(input, dict):
                return self.update(input)
            elif isinstance(input, list):
                return self.extend(input)
            else:
                return self.set(input)
        else:
            current_content = self.get(input)
            if isinstance(current_content, list):
                if isinstance(desc, list):
                    return self.extend(input, desc)
                else:
                    return self.append(input, desc)
            elif isinstance(desc, dict):
                return self.update(input, desc)
            else:
                return self.set(input, desc)

    def set(self, keys_with_dots: Any, value: Any = None) -> 'NamespaceOps':
        """设置键值对"""
        if value is None:
            self.data_ops.set(self.namespace_name, keys_with_dots)
        else:
            self.data_ops.set(f"{self.namespace_name}.{keys_with_dots}", value)
        return self.return_to

    def delta(self, keys_with_dots: str, value: Any) -> None:
        """递归处理差异更新"""
        if isinstance(value, dict):
            for k, v in value.items():
                self.delta(f"{keys_with_dots}.{k}", v)
        else:
            self.append(keys_with_dots, value)

    def append(self, keys_with_dots: Any, value: Any = None) -> 'NamespaceOps':
        """追加值到列表"""
        if value is None:
            self.data_ops.append(self.namespace_name, keys_with_dots)
        else:
            self.data_ops.append(f"{self.namespace_name}.{keys_with_dots}", value)
        return self.return_to

    def extend(self, keys_with_dots: Any, value: Any = None) -> 'NamespaceOps':
        """扩展列表"""
        if value is None:
            self.data_ops.extend(self.namespace_name, keys_with_dots)
        else:
            self.data_ops.extend(f"{self.namespace_name}.{keys_with_dots}", value)
        return self.return_to

    def update(self, keys_with_dots: Any, value: Any = None) -> 'NamespaceOps':
        """更新字典"""
        if value is None:
            self.data_ops.update(self.namespace_name, keys_with_dots)
        else:
            self.data_ops.update(f"{self.namespace_name}.{keys_with_dots}", value)
        return self.return_to
        
    def get(self, keys_with_dots: Optional[str] = None, default: Any = None, *, no_copy: bool = False) -> Any:
        """获取值"""
        key = f"{self.namespace_name}.{keys_with_dots}" if keys_with_dots else self.namespace_name
        return self.data_ops.get(key, default, no_copy=no_copy)

    def remove(self, keys_with_dots: str) -> Any:
        """删除键"""
        return self.data_ops.remove(f"{self.namespace_name}.{keys_with_dots}")

    def empty(self) -> Any:
        """清空命名空间"""
        return self.data_ops.remove(self.namespace_name)


class DataOps:
    def __init__(self, *, target_data: Optional[Dict] = None, no_copy: bool = False):
        self.target_data: Dict = target_data or {}
        self.no_copy = no_copy

    def __locate_pointer(self, keys_with_dots: str) -> tuple:
        """定位到目标字典和最终键"""
        keys = keys_with_dots.split('.')
        pointer = self.target_data
        current_key = None
        
        for key in keys:
            if current_key:
                pointer = pointer[current_key]
            current_key = key
            if key not in pointer:
                pointer[key] = {}
                
        return pointer, current_key

    def set(self, keys_with_dots: str, value: Any) -> 'DataOps':
        """设置值"""
        pointer, key = self.__locate_pointer(keys_with_dots)
        pointer[key] = value
        return self

    def delta(self, keys_with_dots: str, value: Any) -> None:
        """增量更新，递归处理嵌套字典"""
        if isinstance(value, dict):
            for k, v in value.items():
                self.delta(f"{keys_with_dots}.{k}", v)
        else:
            if self.get(keys_with_dots):
                self.append(keys_with_dots, value)
            else:
                self.set(keys_with_dots, value)

    def __prepare_list_for_append(self, pointer: Dict, key: str) -> None:
        """准备列表以供追加"""
        if not isinstance(pointer[key], list):
            if pointer[key] == {}:
                pointer[key] = []
            else:
                pointer[key] = [pointer[key]]

    def append(self, keys_with_dots: str, value: Any) -> 'DataOps':
        """追加值到列表"""
        pointer, key = self.__locate_pointer(keys_with_dots)
        self.__prepare_list_for_append(pointer, key)
        pointer[key].append(value)
        return self

    def extend(self, keys_with_dots: str, value: Any) -> 'DataOps':
        """扩展列表"""
        if not isinstance(value, list):
            value = [value]
            
        pointer, key = self.__locate_pointer(keys_with_dots)
        self.__prepare_list_for_append(pointer, key)
        pointer[key].extend(value)
        return self

    def __update_dict(self, pointer: Dict, pointer_key: str, dict_item: Any) -> None:
        """递归更新字典"""
        if isinstance(dict_item, dict):
            for key in dict_item:
                if key not in pointer[pointer_key]:
                    pointer[pointer_key][key] = {}
                    
                if isinstance(dict_item[key], dict):
                    self.__update_dict(pointer[pointer_key], key, dict_item[key])
                elif dict_item[key] is not None:
                    pointer[pointer_key][key] = dict_item[key]
        elif dict_item is not None:
            pointer[pointer_key] = dict_item

    def update(self, keys_with_dots: str, value: Any) -> 'DataOps':
        """更新字典"""
        pointer, key = self.__locate_pointer(keys_with_dots)
        self.__update_dict(pointer, key, value)
        return self

    def update_by_dict(self, data_dict: Dict) -> 'DataOps':
        """通过字典批量更新"""
        for key, value in data_dict.items():
            self.update(key, value)
        return self

    def _deep_get(self, data: Any) -> Any:
        """深度复制数据，处理特殊类型"""
        if isinstance(data, dict):
            result = {}
            for key, value in data.items():
                result[key] = self._deep_get(value)
            return result
        elif isinstance(data, list):
            return [self._deep_get(item) for item in data]
        elif isinstance(data, set):
            return {self._deep_get(item) for item in data}
        elif isinstance(data, tuple):
            return tuple(self._deep_get(item) for item in data)
        elif hasattr(data, '__class__') and hasattr(data, '__module__') and data.__module__ != 'builtins':
            return data
        else:
            return copy.deepcopy(data)

    def get(self, keys_with_dots: Optional[str] = None, default: Any = None, *, no_copy: bool = False) -> Any:
        """获取值，支持点分隔路径"""
        # 处理空键情况
        if not keys_with_dots:
            return self._handle_copy(self.target_data, no_copy)
            
        # 查找路径
        keys = keys_with_dots.split('.')
        pointer = self.target_data
        
        for key in keys:
            if key not in pointer:
                return default
            pointer = pointer[key]
            
        return self._handle_copy(pointer, no_copy)
    
    def _handle_copy(self, data: Any, no_copy: bool) -> Any:
        """根据复制标志处理数据复制"""
        if self.no_copy or no_copy:
            return data
            
        if hasattr(data, '__class__') and hasattr(data, '__module__') and data.__module__ != 'builtins':
            return data
            
        try:
            return copy.deepcopy(data)
        except TypeError:
            return self._deep_get(data)

    def remove(self, keys_with_dots: str) -> 'DataOps':
        """删除指定键"""
        pointer, key = self.__locate_pointer(keys_with_dots)
        del pointer[key]
        return self

    def empty(self) -> 'DataOps':
        """清空所有数据"""
        self.target_data = {}
        return self


from collections import deque
from typing import List
from ..utils.chunk_helper import deep_copy_simply

class RuntimeBranchState:
  """某个分支的运行时状态"""
  def __init__(self, **args) -> None:
    self.id = args.get('id')
    # 恢复各 chunk 的运行状态
    self.chunk_status = args.get('chunk_status', {})
    self.executing_ids: List[str] = args.get('executing_ids', [])
    self.visited_record: List[str] = args.get('visited_record', [])
    # 总运行状态
    self.running_status = args.get('running_status', 'idle')
    # 快队列
    self.running_queue = deque(args.get('running_queue', []))
    # 慢队列
    self.slow_queue = deque(args.get('slow_queue', []))

  def update_chunk_status(self, chunk_id, status):
    if status not in ['idle', 'success', 'running', 'error', 'pause']:
      return
    self.chunk_status[chunk_id] = status

  def get_chunk_status(self, chunk_id):
    return self.chunk_status.get(chunk_id)
  
  def export(self):
    return deep_copy_simply({
      'id': self.id,
      'executing_ids': self.executing_ids,
      'chunk_status': self.chunk_status,
      'visited_record': self.visited_record,
      'running_status': self.running_status,
      'running_queue': list(self.running_queue),
      'slow_queue': list(self.slow_queue)
    })



from .Stage import Stage

class EventEmitter:
    def __init__(self, private_max_workers=None, max_concurrent_tasks=None, on_error=None):
        self._listeners = {}
        self._once = {}
        self._stage = Stage(
            private_max_workers=private_max_workers,
            max_concurrent_tasks=max_concurrent_tasks,
            on_error=on_error,
            is_daemon=True,
        )
    
    def add_listener(self, event, listener):
        if event not in self._listeners:
            self._listeners.update({ event: [] })
        if listener not in self._listeners[event]:
            self._listeners[event].append(listener)
        return listener
    
    def remove_listener(self, event, listener):
        if event in self._listeners and listener in self._listeners[event]:
            self._listeners[event].remove(listener)

    def remove_all_listeners(self, event_list):
        if isinstance(event_list, str):
            event_list = [event_list]
        for event in event_list:
            self._listeners.update({ event: [] })

    def on(self, event, listener):
        return self.add_listener(event, listener)
    
    def off(self, event, listener):
        return self.remove_listener(event, listener)

    def once(self, event, listener):
        if event not in self._once:
            self._once.update({ event: [] })
        if listener not in self._listeners[event] and listener not in self._once[event]:
            self._once[event].append(listener)
        return listener
    
    def listener_count(self, event):
        return len(self._listeners[event]) + len(self._once[event])
        
    def emit(self, event, *args, **kwargs):
        listeners_to_execute = []
        if event in self._listeners:
            for listener in self._listeners[event]:
                listeners_to_execute.append((listener, args, kwargs))
        if event in self._once:
            for listener in self._once[event]:
                listeners_to_execute.append((listener, args, kwargs))
            self._once.update({ event: [] })
        for listener, args, kwargs in listeners_to_execute:
            self._stage.go(listener, *args, **kwargs)
        return len(listeners_to_execute)


from .lexer import Lexer

__version__ = "0.0.4"

__title__ = "streamingjson"
__description__ = (
    "A streamlined, user-friendly JSON streaming preprocessor, crafted in Python."
)
__url__ = "https://github.com/karminski/streaming-json-py"
__uri__ = __url__
__doc__ = f"{__description__} <{__uri__}>"

__author__ = "Karminski"
__email__ = "code.karminski@outlook.com"

__license__ = "MIT"
__copyright__ = "Copyright 2024 Karminski"

__all__ = [
    "Lexer",
]



import os
import importlib
import configparser

def export():
    plugin_list = []

    dir_path = os.path.dirname(os.path.abspath(__file__))
    dir_name = os.path.basename(dir_path)
    
    # read config.ini
    config = configparser.ConfigParser()
    config.optionxform = str
    config.read(f"{ dir_path }/config.ini")
    config_is_not_empty = len(dict(config)) > 1
    
    # load configs
    module_name = config["plugins"]["module_name"] if config_is_not_empty and "module_name" in config else dir_name
    exception_files = set(config["plugins"]["exception_files"]) if config_is_not_empty and "exception_files" in config["plugins"] else set([])
    exception_files.add("__init__.py")
    exception_dirs = set(config["plugins"]["exception_dirs"]) if config_is_not_empty and "exception_dirs" in config["plugins"] else set([])
    exception_dirs.add("utils")
    default_settings = dict(config["default settings"]) if config_is_not_empty and "default settings" in config else None

    for item in os.listdir(dir_path):
        # import plugins in .py files
        if item.endswith('.py') and item not in exception_files:
            plugin = importlib.import_module(f".{ item[:-3] }", package = __package__)
            if not hasattr(plugin, "export"):
                raise Exception(f"[Plugin Manager] Function 'export' must be stated in plugin file: { dir_path }/{ item }")
            plugin_export = getattr(plugin, "export")()
            plugin_list.append((module_name, plugin_export[0], plugin_export[1]))
        # import plugins in dirs
        if os.path.exists(f"{ dir_path }/{ item }/__init__.py") and item not in exception_dirs:
            plugin = importlib.import_module(f".{ item }", package = __package__)
            if not hasattr(plugin, "export"):
                raise Exception(f"[Plugin Manager] Function 'export' must be stated in plugin dir's '__init__.py' file: { dir_path }/{ item }/__init__.py")
            plugin_export = getattr(plugin, "export")()
            plugin_list.append((module_name, plugin_export[0], plugin_export[1]))

    return (plugin_list, default_settings)


from .utils import PluginManager, StorageDelegate, ToolManager, RuntimeCtx
from .plugins import install_plugins

global_plugin_manager = PluginManager()
global_tool_manager = ToolManager()
global_settings = RuntimeCtx()
install_plugins(global_plugin_manager, global_tool_manager, global_settings)
global_storage = StorageDelegate(
    db_name = "global",
    plugin_manager = global_plugin_manager,
    settings = global_settings,
)


from typing import Dict, TYPE_CHECKING
from .BranchState import RuntimeBranchState

if TYPE_CHECKING:
    from .Snapshot import Snapshot

CURRENT_STATE_VERSION = 1

class RuntimeState:
  """整个 workflow 的运行时状态"""

  def __init__(self, **args) -> None:
    if args.get('v') and args.get('v') != CURRENT_STATE_VERSION:
      raise ValueError(f'The state data of version {args.get("v")} cannot be used on the checkpoint management logic of version {CURRENT_STATE_VERSION}.')
    # schema 版本
    self.v = CURRENT_STATE_VERSION
    self.workflow_id = args.get('workflow_id')
    # 分支决策逻辑
    self.branches_state: Dict[str, RuntimeBranchState] = args.get(
        'branches_state', {})
    # 各 chunk 依赖数据
    self.chunks_dep_state = args.get('chunks_dep_state', {})
    # 各 chunk 附加数据（如 Loop chunk 的 State 数据）
    self.chunks_extra_data = args.get('chunks_extra_data', {})
    # 总运行状态
    self.running_status = args.get('running_status', 'idle')
    # 运行数据
    self.user_store = args.get('user_store')
    self.sys_store = args.get('sys_store')
    self.restore_mode = False
  
  def restore_from_snapshot(self, snapshot: 'Snapshot') -> 'RuntimeState':
    """从指定表述结构中恢复"""
    schema = snapshot.export().get('state')
    # 版本校验
    if schema.get('v') and schema.get('v') != CURRENT_STATE_VERSION:
      raise ValueError(
          f'The state data of version {schema.get("v")} cannot be used on the checkpoint management logic of version {CURRENT_STATE_VERSION}.')
    self.workflow_id = schema.get('workflow_id')

    # 恢复挂载实例化的 branch_state
    branches_state = schema.get('branches_state')
    self.branches_state = {}
    for entry_id in branches_state:
      self.branches_state[entry_id] = RuntimeBranchState(**branches_state[entry_id])
    
    # 恢复依赖数据
    self.chunks_dep_state = schema.get('chunks_dep_state')
    # 恢复运行状态
    self.running_status = schema.get('running_status')
    # 恢复用户 store
    self.user_store = self.user_store.__class__(schema.get('user_store'))
    # 恢复系统 store
    self.sys_store = self.sys_store.__class__(schema.get('sys_store'))
    # 恢复模式
    self.restore_mode = True
    return self

  def update(self, chunks_dep_state: dict = {}):
    # 各 chunk 依赖数据
    self.chunks_dep_state = chunks_dep_state or {}

  def create_branch_state(self, chunk) -> RuntimeBranchState:
    self.branches_state[chunk['id']] = RuntimeBranchState(id=chunk['id'])
    return self.branches_state[chunk['id']]
  
  def get_branch_state(self, chunk) -> RuntimeBranchState:
    return self.branches_state.get(chunk['id'])
  
  def export(self):
    # 将分支状态实例导出状态原值
    branches_state_value = {}
    for id in self.branches_state:
      branches_state_value[id] = self.branches_state[id].export()
    # 返回快照数据
    return {
      'v': self.v,
      'workflow_id': self.workflow_id,
      'branches_state': branches_state_value,
      'chunks_dep_state': self.chunks_dep_state,
      'running_status': self.running_status,
      'user_store': self.user_store.get_all(),
      'sys_store': self.sys_store.get_all()
    }



import yaml
import asyncio
from typing import Union, List
from datetime import datetime
from .utils import ComponentABC
from Agently.utils import RuntimeCtxNamespace, to_json_desc
from Agently.Workflow import Workflow
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client
from ...Stage import Stage

from mcp.server.fastmcp import FastMCP

class Tool(ComponentABC):
    def __init__(self, agent: object):
        self._agent = agent
        self._is_debug = lambda: self._agent.settings.get_trace_back("is_debug")
        self._settings = RuntimeCtxNamespace("plugin_settings.agent_component.Tool", self._agent.settings)
        self._tool_manager = self._agent.tool_manager
        self._tools_info = {}
        self._tools_mapping = {}
        self._must_call_tools_info = None
        self._tool_using_workflow = self._get_default_tool_using_workflow()

    # put_event
    async def _put_event(self, event, data):
        await self._agent.call_event_listeners(event, data)
        self._agent.put_data_to_generator(event, data)

    # Normal Tool
    def register_tool(
        self,
        tool_name: str,
        desc: any,
        args: dict,
        func: callable,
        *,
        categories: Union[str, List[str]] = None,
    ):
        self._tool_manager.register(
            tool_name,
            desc,
            args,
            func,
            categories=categories,
        )
        self._tools_info.update({
            tool_name: {
                "tool_name": tool_name,
                "desc": desc,
                "kwargs": to_json_desc(args),
            }
        })
        return self._agent

    def remove_tools(self, tool_names: Union[str, List[str]]):
        if isinstance(tool_names, str):
            tool_names = [tool_names]
        for tool_name in tool_names:
            if tool_name in self._tools_info:
                del self._tools_info[tool_name]
        return self._agent
    
    def use_public_tools(self, tool_names: Union[str, List[str]]):
        if isinstance(tool_names, str):
            tool_names = [tool_names]
        for tool_name in tool_names:
            tool_info = self._tool_manager.get_tool_info(tool_name, with_args=True)
            if tool_info:
                self._tools_info.update({ tool_name: tool_info })
        return self._agent
    
    def use_public_categories(self, tool_categories: Union[str, List[str]]):
        if isinstance(tool_categories, str):
            tool_categories = [tool_categories]
        tools_info = self._tool_manager.get_tool_dict(categories=tool_categories, with_args=True)
        for key, value in tools_info.items():
            self._tools_info.update({ key: value })
        return self._agent

    def use_all_public_tools(self):
        tools_info = self._tool_manager.get_tool_dict(with_args=True)
        for key, value in tools_info.items():
            self._tools_info.update({ key: value })
        return self._agent
    
    # Must Call
    def must_call(self, tool_name: str):
        tool_info = self._tool_manager.get_tool_info(tool_name, with_args=True)
        if tool_info:
            self._must_call_tools_info = tool_info
        else:
            raise NotImplementedError(f"[Agent Component] Tool-must call: can not find tool named '{ tool_name }'.")
        return self._agent
    
    # MCP
    async def _get_tools_info_from_mcp_async(self, *, mcp_server_params: StdioServerParameters):
        async with stdio_client(mcp_server_params) as (read, write):
            async with ClientSession(read, write) as session:
                tools_info = {}
                await session.initialize()
                tools_data = await session.list_tools()
                for tool in tools_data.tools:
                    tools_info.update({
                        tool.name: {
                            "tool_name": tool.name,
                            "desc": tool.description,
                            "kwargs": {}
                        },
                    })
                    for key, value in tool.inputSchema["properties"].items():
                        desc_str = ""
                        if 'title' in value:
                            desc_str += f"{ value['title'] }\n"
                        if 'description' in value:
                            desc_str += f"{ value['description'] }"
                        tools_info[tool.name]["kwargs"].update({
                            key: (
                                value["type"],
                                desc_str,
                            ),
                        })
        return tools_info

    def _get_tools_info_from_mcp(self, *, mcp_server_params: StdioServerParameters):
        return asyncio.run(self._get_tools_info_from_mcp_async(mcp_server_params=mcp_server_params))

    def _generate_mcp_tool(self, mcp_server_params: StdioServerParameters, tool_name:str):
        async def _call_mcp_tool_async(**kwargs):
            async with stdio_client(mcp_server_params) as (read, write):
                async with ClientSession(read, write) as session:
                    await session.initialize()
                    return await session.call_tool(tool_name, arguments=kwargs)
        def _call_mcp_tool(**kwargs):
            with Stage() as stage:
                result = stage.get(_call_mcp_tool_async, **kwargs)
                if result.isError:
                    return {
                        "Error": vars(result.content[0])
                    }
                else:
                    return vars(result.content[0])
        return _call_mcp_tool
    
    def use_mcp_server(
        self,
        command:str=None,
        args:List[str]=None,
        env:str=None,
        *,
        config:dict=None,
        categories:Union[str, List[str]]=None,
    ):
        if command and args:
            mcp_server_params = StdioServerParameters(
                command=command,
                args=args,
                env=env,
            )
            tools_info = self._get_tools_info_from_mcp(mcp_server_params=mcp_server_params)
            for tool_name, tool_info in tools_info.items():
                self._tools_info.update({ tool_name: tool_info })
                self._tool_manager.register(
                    tool_info["tool_name"],
                    tool_info["desc"],
                    to_json_desc(tool_info["kwargs"]),
                    self._generate_mcp_tool(mcp_server_params, tool_info["tool_name"]),
                    categories=categories,          
                )
        elif config:
            if "mcpServers" in config:
                config = config["mcpServers"]
            for server_name, server_info in config.items():
                if "command" in server_info and "args" in server_info:
                    self.use_mcp_server(
                        command=server_info["command"],
                        args=server_info["args"],
                        env=server_info["env"] if "env" in server_info else None,
                        categories=categories,
                    )
        else:
            raise NotImplementedError(f"[Agent Component] Tool - Use MCP Server: `command`, `arg` or `config` dict must be provided.")
        return self._agent
    
    # Customize Workflow
    def set_tool_using_workflow(self, workflow: Workflow):
        self._tool_using_workflow = workflow
        return self._agent

    # Default Workflow
    def _get_default_tool_using_workflow(self):
        tool_using_workflow = Workflow()

        @tool_using_workflow.chunk()
        def save_user_input(inputs, storage):
            storage.set("user_input", inputs["default"])
            return

        @tool_using_workflow.chunk()
        async def make_next_plan(inputs, storage):
            agent = storage.get("$agent")
            user_input = storage.get("user_input", {})
            tools_info = storage.get("tools_info", {})
            done_plans = storage.get("done_plans", [])
            result = (
                agent
                    .input(user_input)
                    .info({
                        "tool_list": [to_json_desc(item) for item in list(tools_info.values())],
                        "already_dones": done_plans,
                        "now": datetime.now().strftime("%Y-%B-%d %H:%M:%S"),
                    })
                    .instruct([
                        "According {input}, {already_dones} and {HELPFUL INFORMATION.tool_list}, make a next move plan",
                        "If {already_dones} shows that some action repeats for too many times, next step MUST BE 'OUTPUT'",
                        "USE WHAT NATURAL LANGUAGE TONGUE {INPUT} USED AS POSSIBLE.",
                    ])
                    .output({
                        "input_natural_language": ("str", ),
                        "next_step_thinking": ("str", ),
                        "next_step_action": {
                            "type": ("'TOOL USING' | 'OUTPUT'", "MUST IN values provided. If {HELPFUL INFORMATION.already_dones} have enough info to reply {input}, MUST choose 'OUTPUT'"),
                            "tool_using": (
                                {
                                    "tool_name": ("str from {HELPFUL INFORMATION.tool_list.tool_name}", "MUST USE {HELPFUL INFORMATION.tool_list.tool_name} provided tool name ONLY!"),
                                    "purpose": ("str", "Describe the problem need to use this tool to solve"),
                                    "kwargs": ("dict", "MUST FOLLOW {HELPFUL INFORMATION.tool_list.kwargs} format requirement!"),
                                },
                                "if {next_step_action.type} == 'OUTPUT' just output null",
                            ),
                        }
                    })
                    .start()
            )
            await self._put_event("tool:planning", result["next_step_thinking"])
            return result["next_step_action"]

        @tool_using_workflow.chunk()
        async def use_tool(inputs, storage):
            tool_using_info = inputs["default"]["tool_using"]
            if storage.get("print_process"):
                print("[🤔 Need to use tool]: ")
                print("❓ Problem is:", tool_using_info["purpose"])
                print("🪛 Use tool:", tool_using_info["tool_name"])
            tool_result = self._tool_manager.call_tool_func(tool_using_info["tool_name"], **tool_using_info["kwargs"])
            if storage.get("print_process"):
                print("ℹ️ Tool result: ", tool_result[:100], "...")
            tool_using_result = {
                "purpose": tool_using_info["purpose"],
                "tool_name": tool_using_info["tool_name"],
                "result": tool_result,
            }
            done_plans = storage.get("done_plans", [])
            done_plans.append(tool_using_result)
            storage.set("done_plans", done_plans)
            await self._put_event("tool:tool_result", tool_using_result)
            return
    
        @tool_using_workflow.chunk()
        def reply(inputs, storage):
            if storage.get("print_process"):
                print("✅ Tool using Done!")
            return storage.get("done_plans", None)

        (
            tool_using_workflow
                .connect_to("save_user_input")
                .connect_to("make_next_plan")
                .if_condition(lambda return_value, storage: return_value["type"] == "OUTPUT")
                    .connect_to("reply")
                    .connect_to("end")
                .else_condition()
                    .connect_to("use_tool")
                    .connect_to("make_next_plan")
        )
        return tool_using_workflow
    
    async def _prefix(self):
        if self._must_call_tools_info:
            if self._is_debug():
                print(f"[Agent Component] Using Tools: Must call '{ self.must_call_tool_info['tool_name'] }'")
            self.agent.request.request_runtime_ctx.remove("prompt.instruct")
            self.agent.request.request_runtime_ctx.remove("prompt.output")
            return {
                "info": {
                    "function_must_call": {
                        "tool_name": self.must_call_tool_info["tool_name"],
                        "desc": self.must_call_tool_info["desc"],
                        "args": self.must_call_tool_info["args"],
                    },
                },
                "output": {
                    "can_call": ("Boolean", "Is all information especially from {input} above enough to generate all arguments in {function_must_call.args}?"),
                    "args": ("if {can_call} == true, generate args dict according {function_must_call.args} requirement, else leave argument value as null"),                    
                    "question": ("String", "if {can_call}==false, output question for user to collecting enough information."),
                }
            }
        elif len(self._tools_info.keys()) > 0:
            prompt = self._agent.request.request_runtime_ctx.get("prompt", {})
            if len(prompt.keys()) == 1 and "input" in prompt:
                prompt =  yaml.safe_dump(prompt["input"], allow_unicode=True)
            else:
                prompt = yaml.safe_dump(prompt, allow_unicode=True)
            tool_using_result = self._tool_using_workflow.start(
                prompt,
                storage={
                    "$agent": self._agent.worker_request,
                    "tools_info":self._tools_info,
                },
            )["default"]
            if tool_using_result and len(tool_using_result) > 0:
                await self._put_event("tool:done", tool_using_result)            
                return {
                    "info": {
                        "tool_using_result": tool_using_result,
                    },
                    "instruct": [
                        "MUST USE WHAT NATURAL LANGUAGE TONGUE {INPUT} USED TO OUTPUT!",
                        "Response according {tool_using_result}",
                        "summarize key points if possible.",                        
                        "Provide URL for keywords in markdown format if needed.",
                        "If error occurred or not enough information was given, response you don't know honestly unless you are very sure about the answer without information support.",
                    ]
                }
            else:
                return None
        else:
            return None
    
    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": None,
            "alias": {
                "register_tool": { "func": self.register_tool },
                "call_tool": { "func": self._tool_manager.call_tool_func, "return_value": True },
                "set_tool_proxy": { "func": self._tool_manager.set_tool_proxy },
                "add_public_tools": { "func": self.use_public_tools },
                "use_public_tools": { "func": self.use_public_tools },
                "add_public_categories": { "func": self.use_public_categories },
                "use_public_categories": { "func": self.use_public_categories },
                "add_all_public_tools": { "func": self.use_all_public_tools },
                "use_all_public_tools": { "func": self.use_all_public_tools },
                "use_mcp_server": { "func": self.use_mcp_server },
                "set_tool_using_workflow": { "func": self.set_tool_using_workflow },
                "remove_tools": { "func": self.remove_tools },
                "must_call": { "func": self.must_call },
            }
        }

def export():
    return ("Tool", Tool)


import queue
import threading
from .Stage import Stage

class Tunnel:
    def __init__(self, private_max_workers=1, max_concurrent_tasks=None, on_error=None, timeout=10):
        self._private_max_worker = private_max_workers
        self._max_concurrent_tasks = max_concurrent_tasks
        self._on_error = on_error
        self._timeout = timeout
        self._data_queue = queue.Queue()
        self._close_event = threading.Event()
        self._stage = None
    
    def _defer_close_stage(self):
        def close_stage():
            self._close_event.wait()
            if self._stage is not None:
                self._stage.close()
                self._stage = None
        defer_thread = threading.Thread(target=close_stage)
        defer_thread.start()
    
    def _get_stage(self):
        if self._stage is not None:
            return self._stage
        else:
            self._stage = Stage(private_max_workers=self._private_max_worker, max_concurrent_tasks=self._max_concurrent_tasks, on_error=self._on_error)
            return self._stage
    
    def put(self, data):
        self._data_queue.put(data)
    
    def put_stop(self):
        self._data_queue.put(StopIteration)
    
    def get(self):
        self._defer_close_stage()
        stage = self._get_stage()
        def queue_consumer():
            try_count = 0
            while True:
                try:
                    if self._timeout is not None:
                        data = self._data_queue.get(timeout=self._timeout)
                    else:
                        data = self._data_queue.get()
                except queue.Empty:
                    if self._timeout is not None and try_count < 3:
                        try_count += 1
                        continue
                    break
                if data is StopIteration:
                    break
                yield data
            self._close_event.set()
        return stage.go(queue_consumer)


from .utils import RequestABC, to_prompt_structure, to_instruction, to_json_desc, find_json, format_request_messages
from Agently.utils import RuntimeCtxNamespace
import json
import boto3

class Bedrock(RequestABC):
    def __init__(self, request):
        self.request = request
        self.model_name = "Bedrock"
        self.model_settings = RuntimeCtxNamespace(f"model.{self.model_name}", self.request.settings)
        
        # Set default message rules
        if not self.model_settings.get_trace_back("message_rules.no_multi_system_messages"):
            self.model_settings.set("message_rules.no_multi_system_messages", False)
        if not self.model_settings.get_trace_back("message_rules.strict_orders"):
            self.model_settings.set("message_rules.strict_orders", True)
        if not self.model_settings.get_trace_back("message_rules.no_multi_type_messages"):
            self.model_settings.set("message_rules.no_multi_type_messages", False)
            
        # Default options for Bedrock models
        self.default_options = {
            "modelId": "anthropic.claude-3-sonnet-20240229-v1:0",
            "maxTokens": 4096,
        }
        
        # For streaming responses
        self.is_streaming = True

    def construct_request_messages(self):
        # Init request messages
        request_messages = []
        
        # General instruction
        general_instruction_data = self.request.request_runtime_ctx.get_trace_back("prompt.general")
        if general_instruction_data:
            request_messages.append({"role": "system", "content": [{"type": "text", "text": f"[GENERAL INSTRUCTION]\n{to_instruction(general_instruction_data)}"}]})
        
        # Role
        role_data = self.request.request_runtime_ctx.get_trace_back("prompt.role")
        if role_data:
            request_messages.append({"role": "system", "content": [{"type": "text", "text": f"[ROLE SETTINGS]\n{to_instruction(role_data)}"}]})
        
        # User info
        user_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.user_info")
        if user_info_data:
            request_messages.append({"role": "system", "content": [{"type": "text", "text": f"[ABOUT USER]\n{to_instruction(user_info_data)}"}]})
        
        # Abstract
        headline_data = self.request.request_runtime_ctx.get_trace_back("prompt.abstract")
        if headline_data:
            request_messages.append({"role": "assistant", "content": [{"type": "text", "text": to_instruction(headline_data)}]})
        
        # Chat history
        chat_history_data = self.request.request_runtime_ctx.get_trace_back("prompt.chat_history")
        if chat_history_data:
            request_messages.extend(chat_history_data)
        
        # Request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get_trace_back("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get_trace_back("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get_trace_back("prompt.output")
        
        # Files
        files_data = self.request.request_runtime_ctx.get_trace_back("prompt.files")
        
        # Validate that at least one prompt field is set
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception("[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        
        prompt_text = ""
        # Only input
        if prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            prompt_text = to_instruction(prompt_input_data)
        # Construct structured prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[INPUT]"] = to_instruction(prompt_input_data)
            if prompt_info_data:
                prompt_dict["[HELPFUL INFORMATION]"] = str(prompt_info_data)
            if prompt_instruct_data:
                prompt_dict["[INSTRUCTION]"] = to_instruction(prompt_instruct_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[OUTPUT REQUIREMENT]"] = {
                        "TYPE": "JSON can be parsed in Python",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[OUTPUT REQUIREMENT]"] = str(prompt_output_data)
            
            prompt_text = to_prompt_structure(prompt_dict, end="[OUTPUT]:\n")
        
        request_messages.append({"role": "user", "content": [{"type": "text", "text": prompt_text}]})
        return request_messages

    def _create_client(self):
        """Create and return a Bedrock Runtime client using AWS credentials."""
        aws_region = self.model_settings.get_trace_back("region", "us-east-1")
        aws_access_key = self.model_settings.get_trace_back("auth.access_key")
        aws_secret_key = self.model_settings.get_trace_back("auth.secret_key")
        aws_session_token = self.model_settings.get_trace_back("auth.session_token", None)
        
        if not aws_access_key or not aws_secret_key:
            raise Exception("[Bedrock] Missing AWS credentials. Use Agently.set_settings('model.Bedrock.auth', {'access_key': '<YOUR_ACCESS_KEY>', 'secret_key': '<YOUR_SECRET_KEY>'}) to set them.")
        
        credentials = {
            "aws_access_key_id": aws_access_key,
            "aws_secret_access_key": aws_secret_key,
        }
        
        if aws_session_token:
            credentials["aws_session_token"] = aws_session_token
            
        client = boto3.client(
            'bedrock-runtime',
            region_name=aws_region,
            **credentials
        )
        
        return client

    def generate_request_data(self):
        options = self.model_settings.get_trace_back("options", {})
        self.is_streaming = options.get("streaming", True)
        
        # Set streaming default
        if "streaming" not in options:
            options["streaming"] = self.is_streaming
            
        # Merge with default options
        for key, value in self.default_options.items():
            if key not in options:
                options[key] = value
        
        return {
            "messages": format_request_messages(self.construct_request_messages(), self.model_settings),
            "options": options,
        }

    async def request_model(self, request_data: dict):
        client = self._create_client()
        messages = request_data["messages"]
        options = request_data["options"]
        model_id = options.pop("modelId")
        is_streaming = options.pop("streaming") if "streaming" in options else self.is_streaming
        
        # Extract system message if present
        system_prompt = ""
        request_messages = []
        
        for message in messages:
            if message["role"] == "system":
                system_prompt += f"{message['content'][0]['text']}\n"
            else:
                request_messages.append(message)
        
        # Prepare the request body based on the model provider
        if "anthropic.claude" in model_id:
            # Claude format for Bedrock
            body = {
                "anthropic_version": "bedrock-2023-05-31",
                "max_tokens": options.pop("maxTokens") if "maxTokens" in options else 4096,
                "messages": request_messages
            }
            
            if system_prompt:
                body["system"] = system_prompt.strip()
                
            # Add remaining options
            for key, value in options.items():
                body[key] = value
                
        elif "amazon.titan" in model_id:
            # Titan format for Bedrock
            body = {
                "inputText": messages[-1]["content"][0]["text"],
                "textGenerationConfig": options
            }
            
        elif "meta.llama" in model_id:
            # Llama format for Bedrock
            combined_prompt = ""
            if system_prompt:
                combined_prompt += f"<system>\n{system_prompt.strip()}\n</system>\n\n"
                
            for msg in request_messages:
                role = msg["role"]
                content = msg["content"][0]["text"]
                if role == "user":
                    combined_prompt += f"<human>\n{content}\n</human>\n\n"
                elif role == "assistant":
                    combined_prompt += f"<assistant>\n{content}\n</assistant>\n\n"
            
            # Add final assistant tag for the model response
            combined_prompt += "<assistant>\n"
            
            body = {
                "prompt": combined_prompt,
                "max_gen_len": options.pop("maxTokens") if "maxTokens" in options else 4096,
            }
            
            # Add remaining options
            for key, value in options.items():
                body[key] = value
                
        else:
            # Generic format as fallback
            body = {
                "prompt": messages[-1]["content"][0]["text"],
                **options
            }
        
        # Convert body to JSON string
        body_json = json.dumps(body)
        
        if is_streaming:
            # Streaming response
            response = client.invoke_model_with_response_stream(
                modelId=model_id,
                body=body_json
            )
            
            for event in response.get('body'):
                if 'chunk' in event:
                    chunk_data = json.loads(event['chunk']['bytes'].decode())
                    yield chunk_data
        else:
            # Non-streaming response
            response = client.invoke_model(
                modelId=model_id,
                body=body_json
            )
            
            yield json.loads(response.get('body').read())

    async def broadcast_response(self, response_generator):
        if not self.is_streaming:
            # Non-streaming response
            async for response in response_generator:
                if "anthropic.claude" in self.default_options["modelId"]:
                    result = response.get("content", [{"text": ""}])[0]["text"]
                elif "amazon.titan" in self.default_options["modelId"]:
                    result = response.get("results", [{"outputText": ""}])[0]["outputText"]
                elif "meta.llama" in self.default_options["modelId"]:
                    result = response.get("generation", "")
                else:
                    result = response.get("completion", "")
                    
                yield({"event": "response:done_origin", "data": response})
                yield({"event": "response:done", "data": result})
        else:
            # Streaming response
            full_response = ""
            
            if "anthropic.claude" in self.default_options["modelId"]:
                async for chunk in response_generator:
                    yield({"event": "response:delta_origin", "data": chunk})
                    
                    if "type" in chunk and chunk["type"] == "content_block_delta":
                        delta = chunk["delta"].get("text", "")
                        if delta:
                            yield({"event": "response:delta", "data": delta})
                            full_response += delta
                    
                    if "type" in chunk and chunk["type"] == "message_stop":
                        yield({"event": "response:done_origin", "data": chunk})
                        yield({"event": "response:done", "data": full_response})
            else:
                # Generic streaming handler for other models
                async for chunk in response_generator:
                    yield({"event": "response:delta_origin", "data": chunk})
                    
                    if "amazon.titan" in self.default_options["modelId"]:
                        delta = chunk.get("outputText", "")
                    elif "meta.llama" in self.default_options["modelId"]:
                        delta = chunk.get("generation", "")
                    else:
                        delta = chunk.get("completion", "")
                        
                    if delta:
                        yield({"event": "response:delta", "data": delta})
                        full_response += delta
                
                yield({"event": "response:done", "data": full_response})

    def export(self):
        return {
            "generate_request_data": self.generate_request_data,
            "request_model": self.request_model,
            "broadcast_response": self.broadcast_response,
        }

def export():
    return ("Bedrock", Bedrock)


from .DataOps import DataOps, NamespaceOps

class RuntimeCtxNamespace(NamespaceOps):
    def __init__(self, namespace_name: str, runtime_ctx: object, *, return_to: object=None):
        super().__init__(namespace_name, runtime_ctx, return_to = return_to)

    def get_trace_back(self, keys_with_dots: str=None, default: str=None):
        return self.data_ops.get_trace_back(f"{ self.namespace_name }.{ keys_with_dots }" if keys_with_dots else self.namespace_name, default)

    def get(self, keys_with_dots: str=None, default: str=None, *, trace_back=True):
        if trace_back:
            return self.get_trace_back(keys_with_dots, default)
        else:
            return self.data_ops.get(f"{ self.namespace_name }.{ keys_with_dots }" if keys_with_dots else self.namespace_name, default)

class RuntimeCtx(DataOps):
    def __init__ (self, *, parent: object=None, no_copy: bool=False):
        self.parent = parent
        self.runtime_ctx_storage = {}
        super().__init__(target_data = self.runtime_ctx_storage, no_copy = no_copy)

    def __update_trace_back_result(self, parent_result, result):
        for key in result:
            if key not in parent_result:
                parent_result[key] = {}
            elif not isinstance(parent_result[key], dict):
                parent_result[key] = {}
            if isinstance(result[key], dict):
                self.__update_trace_back_result(parent_result[key], result[key])
            else:
                parent_result[key] = result[key]
        return parent_result

    def get_trace_back(self, keys_with_dots:str=None, default: str=None):
        result = self.get(keys_with_dots, trace_back = False)
        parent_result = self.parent.get_trace_back(keys_with_dots) if self.parent else None
        if result or parent_result:
            if isinstance(result, dict):
                parent_result = parent_result if isinstance(parent_result, dict) else {}
                return self.__update_trace_back_result(parent_result, result)
            else:
                return result if result is not None else parent_result
        else:
            return default

    def get(self, keys_with_dots: str=None, default: str=None, *, trace_back=True):
        if trace_back:
            return self.get_trace_back(keys_with_dots, default)
        else:
            return super().get(keys_with_dots, default)


import re
from .utils import ToolABC
import time
import httpx
from duckduckgo_search import DDGS
import requests
from bs4 import BeautifulSoup

class Web(ToolABC):
    def __init__(self, tool_manager: object):
        self.tool_manager = tool_manager

    def search(self, keywords: str, *, options: dict={}, proxy: str=None, type: int=1, no_sleep: bool=False, timelimit: str=None):
        if not isinstance(options, dict):
            options = {}
        if timelimit:
            options.update({ "timelimit": timelimit })
        results = []
        '''
        results = {
            "origin": [],
            "for_agent": [],
        }
        '''
        search_kwargs = {}
        if proxy:
            search_kwargs["proxy"] = proxy
        if "max_results" not in options:
            options.update({ "max_results": 5 })
        try:
            with DDGS(**search_kwargs) as ddgs:
                if type == 1:
                    for index, result in enumerate(ddgs.text(
                        str(keywords),
                        **options
                    )):
                        results.append({
                            "title": result["title"],
                            "brief": result["body"],
                            "url": result["href"],
                        })
                elif type == 2:
                    for index, result in enumerate(ddgs.news(
                        str(keywords),
                        **options
                    )):
                        results.append({
                            "title": result["title"],
                            "brief": result["body"],
                            "url": result["url"],
                            "source": result["source"],
                            "date": result["date"],
                        })
        except Exception as e:
            results = f"No Result Because: { str(e) }"
        return results

    def search_definition(self, item_name: str, *, options: dict={}, proxy: str=None):
        return self.search(f'whatis { item_name }', options=options, proxy=proxy)

    def search_news(self, keywords: str, *, options: dict={}, proxy: str=None, timelimit: str=None):
        if timelimit:
            options.update({ "timelimit": timelimit })
        if "timelimit" not in options:
            options.update({ "timelimit": "w" })
        if "max_results" not in options:
            options.update({ "max_results": 10 })
        return self.search(keywords, options=options, proxy=proxy, type=2)

    def browse(self, url: str, *, retry_times:int = 0, proxy: str=None):
        content = ""
        try:
            request_options = {
                "headers": { "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36" }
            }
            if proxy:
                if proxy.startswith("http:"):
                    request_options.update({ "proxies": { "http": proxy } })
                elif proxy.startswith("https:"):
                    request_options.update({ "proxies": { "https": proxy } })
            page = requests.get(
                url,
                **request_options
            )
            soup = BeautifulSoup(page.content, "html.parser")
            # find text in p, list, pre (github code), td
            chunks = soup.find_all(["h1", "h2", "h3", "h4", "h5", "p", "pre", "td"])
            for chunk in chunks:
                if chunk.name.startswith("h"):
                    content += "#" * int(chunk.name[-1]) + " " + chunk.get_text() + "\n"
                else:
                    text = chunk.get_text()
                    if text and text != "":
                        content += text + "\n"
            # find text in div that class=content
            divs = soup.find("div", class_="content")
            if divs:
                chunks_with_text = divs.find_all(text=True)
                for chunk in chunks_with_text:
                    if isinstance(chunk, str) and chunk.strip():
                        content += chunk.strip() + "\n"
            content = re.sub(r"\n+", "\n", content)
            return content
        except Exception as e:
            return f"Can not browse '{ url }'.\tError: { str(e) }"

    def export(self):
        return {
            "search": {
                "desc": "search {keywords}.",
                "args": {
                    "keywords": ("String", "[*Required]keywords to search, seperate keywords by ' '."),
                    "options": {
                        "timelimit": ("String | Null", "'d': last day; 'w': this week; 'm': this month;")
                    },
                },
                "func": self.search,
                "require_proxy": True,
            },
            "search_info": {
                "desc": "search information about {keywords}.",
                "args": {
                    "keywords": ("String", "[*Required]keywords to search, seperate keywords by ' '."),
                    "options": {
                        "timelimit": ("String | Null", "'d': last day; 'w': this week; 'm': this month;")
                    },
                },
                "func": self.search,
                "require_proxy": True,
            },
            "search_news": {
                "desc": "search news about {keywords}.",
                "args": {
                    "keywords": ("String", "[*Required]keywords to search, seperate keywords by ' '."),
                    "options": {
                        "timelimit": ("String | Null", "'d': last day; 'w': this week; 'm': this month;")
                    },
                },
                "func": self.search_news,
                "require_proxy": True,
            },
            "search_definition": {
                "desc": "search definition about {item_name}.",
                "args": {
                    "item_name": ("String", "[*Required]item name to find the definition."),
                },
                "func": self.search_definition,
                "require_proxy": True,
            },
            "browse": {
                "desc": "Browse web page by url.",
                "args": {
                    "url": ("String", "[*Required]url to browse.")
                },
                "func": self.browse,
                "require_proxy": True,
            }
        }

def export():
    return ("Web", Web)


import atexit
import inspect
import threading
import asyncio
from concurrent.futures import ThreadPoolExecutor, Future
from .StageResponse import StageResponse
from .StageHybridGenerator import StageHybridGenerator
from .StageFunction import StageFunctionMixin

class BaseStage:
    _global_executor = None
    _global_max_workers = 5
    _executor_lock = threading.RLock()

    @staticmethod
    def _get_global_executor():
        if BaseStage._global_executor is None:
            with BaseStage._executor_lock:
                BaseStage._global_executor = ThreadPoolExecutor(max_workers=BaseStage._global_max_workers)
        return BaseStage._global_executor
        
    @staticmethod
    def set_global_max_workers(global_max_workers):
        BaseStage._global_max_workers = global_max_workers
        with BaseStage._executor_lock:
            if BaseStage._global_executor:
                BaseStage._global_executor.shutdown(wait=True)
                BaseStage._global_executor = ThreadPoolExecutor(max_workers=BaseStage._global_max_workers)
                atexit.register(BaseStage._global_executor.shutdown)
    
    def __init__(self, private_max_workers=5, max_concurrent_tasks=None, on_error=None, is_daemon=False):
        self._private_max_workers = private_max_workers
        self._max_concurrent_tasks = max_concurrent_tasks
        self._on_error = on_error
        self._is_daemon = is_daemon
        self._semaphore = None
        self._loop_thread = None
        self._loop = None
        self._current_executor = None
        self._loop_ready = threading.Event()
        self._responses = set()
        self._closed = False
        if self._is_daemon:
            atexit.register(self.close)
        self._initialize()
    
    def __enter__(self):
        self._initialize()
        return self
    
    def __exit__(self, type, value, traceback):
        self.close()
        if type is not None and self._on_error is not None:
            self._on_error(value)
        return False

    @property
    def _executor(self):
        if self._current_executor is not None:
            return self._current_executor
        if self._private_max_workers:
            self._current_executor = ThreadPoolExecutor(max_workers=self._private_max_workers)
            return self._current_executor
        else:
            self._current_executor = BaseStage._get_global_executor()
            return self._current_executor
    
    def _initialize(self):
        self._closed = False
        if (
            not self._loop_thread
            or not self._loop_thread.is_alive()
            or not self._loop
            or not self._loop.is_running()
        ):
            self._loop_thread = threading.Thread(target=self._start_loop, daemon=self._is_daemon)
            self._loop_thread.start()
            self._loop_ready.wait()

    def _start_loop(self):
        self._loop = asyncio.new_event_loop()
        self._loop.set_exception_handler(self._loop_exception_handler)
        if self._max_concurrent_tasks:
            self._semaphore = asyncio.Semaphore(self._max_concurrent_tasks)
        asyncio.set_event_loop(self._loop)
        self._loop.call_soon(lambda: self._loop_ready.set())
        self._loop.run_forever()
    
    def _loop_exception_handler(self, loop, context):
        if self._on_error is not None:
            loop.call_soon_threadsafe(self._on_error, context["exception"])
        else:
            raise context["exception"]
    
    def go(self, task, *args, on_success=None, on_error=None, lazy=False, async_gen_interval=0.1, **kwargs):
        if not self._loop or self._loop.is_running():
            self._initialize()
        response_kwargs = {
            "on_success": on_success,
            "on_error": on_error,
        }
        hybrid_generator_kwargs = {
            "on_success": on_success,
            "on_error": on_error,
            "lazy": lazy,
            "async_gen_interval": async_gen_interval,
        }
        if inspect.iscoroutine(task):
            if self._semaphore:
                if not self._semaphore.locked():
                    return StageResponse(self, task, **response_kwargs)
                else:
                    self._semaphore.acquire()
                    try:
                        return StageResponse(self, task, **response_kwargs)
                    finally:
                        self._semaphore.release()
            else:
                return StageResponse(self, task, **response_kwargs)
        elif inspect.isasyncgen(task):
            if self._semaphore:
                if not self._semaphore.locked():
                    return StageHybridGenerator(self, task, **hybrid_generator_kwargs)
                else:
                    self._semaphore.acquire()
                    try:
                        return StageHybridGenerator(self, task, **hybrid_generator_kwargs)
                    finally:
                        self._semaphore.release()
            else:
                return StageHybridGenerator(self, task, **hybrid_generator_kwargs)
        elif asyncio.isfuture(task):
            return StageResponse(self, task, **response_kwargs)
        elif inspect.isgenerator(task):
            async def async_generator():
                for item in task:
                    result = await self._loop.run_in_executor(self._executor, lambda: item) if not self._executor._shutdown else item
                    yield result
            return self.go(async_generator())
        elif inspect.iscoroutinefunction(task) or inspect.isasyncgenfunction(task):
            return self.go(task(*args, **kwargs), **hybrid_generator_kwargs)
        elif inspect.isgeneratorfunction(task):
            return self.go(task(*args, **kwargs), **hybrid_generator_kwargs)
        elif inspect.isfunction(task) or inspect.ismethod(task):
            if not self._executor._shutdown:
                return StageResponse(self, self._loop.run_in_executor(self._executor, lambda: task(*args, **kwargs)), **response_kwargs)
            else:
                try:
                    future = Future()
                    return StageResponse(self, future, **response_kwargs)
                finally:
                    try:
                        future.set_result(task(*args, **kwargs))
                    except Exception as e:
                        future.set_exception(e)
        else:
            raise TypeError(f"Task seems like a value or an executed function not an executable task: { task }")
    
    def go_all(self, *task_list):
        response_list = []
        for task in task_list:
            if isinstance(task, (tuple, list)):
                func, *args, kwargs = task
                response = self.go(func, *args, **kwargs)
            else:
                response = self.go(task)
            response_list.append(response)
        return response_list

    def get(self, task, *args, **kwargs):
        response = self.go(task, *args, **kwargs)
        if isinstance(response, StageResponse):
            return response.get()
        elif isinstance(response, StageHybridGenerator):
            result = []
            for item in response:
                result.append(item)
            return result
    
    def get_all(self, *task_list):
        result_list = []
        for task in task_list:
            if isinstance(task, (tuple, list)):
                func, *args, kwargs = task
                response = self.go(func, *args, **kwargs)
            else:
                response = self.go(task)
            if isinstance(response, StageResponse):
                result_list.append(response.get())
            else:
                result_list.append(response)
        return result_list

    def on_error(self, handler):
        self._on_error = handler
    
    def close(self):
        if self._closed:
            return
        self._closed = True
        
        for response in self._responses.copy():
            response.get()
        
        if self._loop and self._loop.is_running():
            self._loop.call_soon_threadsafe(self._loop.stop)
        if self._loop:
            pending = asyncio.all_tasks(self._loop)
            if pending:
                self._loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))
        if self._private_max_workers and self._current_executor is not None:
            self._current_executor.shutdown(wait=True)
            self._current_executor = None
        if self._loop_thread and self._loop_thread.is_alive():
            self._loop_thread.join()
            self._loop_thread = None
        if self._loop and not self._loop.is_closed():
            self._loop.close()
        self._loop = None

class Stage(BaseStage, StageFunctionMixin):
    pass


import setuptools

with open('./Agently/requirements.txt') as f:
    origin_requirements = f.read().splitlines()

requirements = []
for requirement in origin_requirements:
    if not requirement.startswith("#"):
        requirements.append(requirement)

setuptools.setup(
    name = "Agently",
    version = "3.5.1.2",
    author = "Maplemx",
    author_email = "maplemx@gmail.com",
    description = "Agently, a framework to build applications based on language model powered intelligent agents.",
    long_description = "https://github.com/Maplemx/Agently",
    url = "https://github.com/Maplemx/Agently",
    license='Apache License, Version 2.0',
    packages = setuptools.find_packages(),
    package_data = {"": ["*.txt", "*.ini"]},
    install_requires= requirements,
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3',
)



from typing import Any, Callable, Optional, List, Dict

class Agent:
    def __init__(
        self,
        *,
        agent_id: Optional[str] = ...,
        auto_save: bool = ...,
        parent_agent_runtime_ctx: Any = ...,
        parent_tool_manager: Any = ...,
        global_storage: Any = ...,
        parent_plugin_manager: Any = ...,
        parent_settings: Any = ...,
        is_debug: Optional[bool] = ...,
    ): ...
    def refresh_plugins(self) -> None: ...
    def toggle_auto_save(self, is_enabled: bool) -> "Agent": ...
    def toggle_component(self, component_name: str, is_enabled: bool) -> "Agent": ...
    def save(self) -> "Agent": ...
    def set_settings(self, settings_key: str, settings_value: Any) -> "Agent": ...
    def set_global_variable(
        self, variable_name: str, variable_value: Any
    ) -> "Agent": ...
    def set_agent_prompt(self, key: str, value: Any) -> "Agent": ...
    def get_agent_prompt(self, key: str) -> Any: ...
    def remove_agent_prompt(self, key: str) -> "Agent": ...
    def register_agent_component(
        self, agent_component_name: str, agent_component_class: Callable
    ) -> "Agent": ...
    def set_agent_component_settings(
        self, agent_component_name: str, key: str, value: Any
    ) -> "Agent": ...
    def attach_workflow(self, name: str, workflow: Any) -> "Agent": ...
    async def start_async(
        self, request_type: Optional[str] = ..., *, return_generator: bool = ...
    ) -> Any: ...
    def start(
        self, request_type: Optional[str] = ..., *, return_generator: bool = ...
    ) -> Any: ...

    # ==== alias_manager 动态注入的常用方法 ====
    def general(self, *args, **kwargs) -> "Agent": ...
    def role(self, *args, **kwargs) -> "Agent": ...
    def user_info(self, *args, **kwargs) -> "Agent": ...
    def abstract(self, *args, **kwargs) -> "Agent": ...
    def chat_history(self, list) -> "Agent": ...
    def input(self, *args, **kwargs) -> "Agent": ...
    def info(self, *args, **kwargs) -> "Agent": ...
    def instruct(self, *args, **kwargs) -> "Agent": ...
    def output(self, *args, **kwargs) -> "Agent": ...
    def file(self, *args, **kwargs) -> "Agent": ...
    def files(self, *args, **kwargs) -> "Agent": ...
    def set_request_prompt(self, key: str, value: Any) -> "Agent": ...
    def get_request_prompt(self, key: str) -> "Agent": ...
    def remove_request_prompt(self, key: str) -> "Agent": ...

    # ==== plugins/agent_component 下所有 alias ====
    # Decorator
    def auto_func(self, func: Callable) -> Callable: ...
    def on_event(self, event: str = ..., *, is_await: bool = ...) -> Callable: ...
    def tool(self, **tool_info_kwrags) -> Callable: ...

    # EventListener
    def add_event_listener(
        self,
        event: str,
        listener: Callable,
        *,
        is_await: bool = ...,
        is_agent_event: bool = ...,
    ) -> "Agent": ...
    def on_delta(
        self, listener: Callable, *, is_await: bool = ..., is_agent_event: bool = ...
    ) -> "Agent": ...
    def on_done(
        self, listener: Callable, *, is_await: bool = ..., is_agent_event: bool = ...
    ) -> "Agent": ...
    def on_finally(
        self, listener: Callable, *, is_await: bool = ..., is_agent_event: bool = ...
    ) -> "Agent": ...
    def on_instant(
        self, listener: Callable, *, is_await: bool = ..., is_agent_event: bool = ...
    ) -> "Agent": ...
    def call_event_listeners(self, event: str, data: Any) -> Any: ...

    # Instant
    def use_instant(self) -> "Agent": ...
    def use_realtime(self) -> "Agent": ...

    # ReplyReformer
    def reform_reply(self, reform_handler: Callable) -> "Agent": ...

    # ResponseGenerator
    def put_data_to_generator(self, event: Any, data: Any) -> None: ...
    def get_generator(self) -> Any: ...
    def get_delta_generator(self) -> Any: ...
    def get_instant_generator(self) -> Any: ...
    def get_realtime_generator(self) -> Any: ...
    def get_instant_keys_generator(self, keys: Any) -> Any: ...
    def get_complete_generator(self) -> Any: ...

    # Role
    def set_role_id(self, role_id: str) -> "Agent": ...
    def set_role(self, key: Any, value: Any = ...) -> "Agent": ...
    def update_role(self, key: Any, value: Any = ...) -> "Agent": ...
    def append_role(self, key: Any, value: Any = ...) -> "Agent": ...
    def extend_role(self, key: Any, value: Any = ...) -> "Agent": ...
    def save_role(self, role_id: str = ...) -> "Agent": ...
    def load_role(self, role_id: str) -> "Agent": ...

    # Search
    def use_search(self) -> "Agent": ...

    # Segment
    def segment(
        self,
        name: str,
        output_prompt: Any,
        listener: Callable = ...,
        *,
        is_streaming: bool = ...,
        is_await: bool = ...,
    ) -> "Agent": ...
    def add_segment_listener(
        self,
        name: str,
        listener: Callable,
        *,
        is_streaming: bool = ...,
        is_await: bool = ...,
    ) -> "Agent": ...
    def on_segment_delta(
        self, name: str, listener: Callable, *, is_await: bool = ...
    ) -> "Agent": ...
    def on_segment_done(
        self, name: str, listener: Callable, *, is_await: bool = ...
    ) -> "Agent": ...

    # Session
    def active_session(self, session_id: str = ...) -> str: ...
    def activate_session(self, session_id: str = ...) -> str: ...
    def save_session(self) -> "Agent": ...
    def stop_session(self) -> "Agent": ...
    def toggle_session_auto_save(self, is_enabled: bool) -> "Agent": ...
    def toggle_strict_orders(self, is_strict_orders: bool) -> "Agent": ...
    def toggle_manual_chat_history(self, is_manual_chat_history: bool) -> "Agent": ...
    def set_chat_history_max_length(self, max_length: int) -> "Agent": ...
    def add_chat_history(self, role: str, content: str) -> "Agent": ...
    def get_chat_history(self, *, is_shorten: bool = ...) -> Any: ...
    def rewrite_chat_history(self, new_chat_history: list) -> "Agent": ...
    def set_abstract(self, *args, **kwargs) -> "Agent": ...

    # Status
    def set_status(self, key: str, value: str) -> "Agent": ...
    def save_status(self) -> "Agent": ...
    def load_status(self) -> "Agent": ...
    def use_global_status(self, namespace_name: str = ...) -> "Agent": ...
    def append_status_mapping(
        self, status_key: str, status_value: str, alias_name: str, *args, **kwargs
    ) -> "Agent": ...

    # Tool
    def register_tool(
        self,
        tool_name: str,
        desc: Any,
        args: dict,
        func: Callable,
        *,
        categories: Any = ...,
    ) -> "Agent": ...
    def call_tool(self, *args, **kwargs) -> Any: ...
    def set_tool_proxy(self, *args, **kwargs) -> "Agent": ...
    def add_public_tools(self, tool_names: Any) -> "Agent": ...
    def use_public_tools(self, tool_names: Any) -> "Agent": ...
    def add_public_categories(self, tool_categories: Any) -> "Agent": ...
    def use_public_categories(self, tool_categories: Any) -> "Agent": ...
    def add_all_public_tools(self) -> "Agent": ...
    def use_all_public_tools(self) -> "Agent": ...
    def use_mcp_server(
        self,
        command: str = ...,
        args: List[str] = ...,
        env: str = ...,
        *,
        config: dict = ...,
        categories: Any = ...,
    ) -> "Agent": ...
    def set_tool_using_workflow(self, workflow: Any) -> "Agent": ...
    def remove_tools(self, tool_names: Any) -> "Agent": ...
    def must_call(self, tool_name: str) -> "Agent": ...

    # UserInfo
    def set_user_name(self, name: str) -> "Agent": ...
    def set_user_info(self, key: Any, value: Any = ...) -> "Agent": ...
    def update_user_info(self, key: Any, value: Any = ...) -> "Agent": ...
    def append_user_info(self, key: Any, value: Any = ...) -> "Agent": ...
    def extend_user_info(self, key: Any, value: Any = ...) -> "Agent": ...
    def save_user_info(self, role_name: str = ...) -> "Agent": ...
    def load_user_info(self, role_name: str) -> "Agent": ...

    # YAMLLoader
    def load_yaml_prompt(
        self,
        *,
        path: str = ...,
        yaml: str = ...,
        use_agently_style: bool = ...,
        variables: dict = ...,
    ) -> "Agent": ...
